"""
Comprehensive test coverage for FileDropResolver.

Test coverage includes:
- Initialization and widget creation
- User API exposure
- Validity checks
- Label generation from file names
- File parsing for various formats (CSV, ASCII, FITS, VOTable)
- Progress tracking
- File upload callbacks
- Multiple file handling
- Edge cases and error handling
"""
import io
import pytest
from unittest.mock import Mock, patch

from astropy.table import Table

from jdaviz.core.loaders.resolvers.file_drop.file_drop import (FileDropResolver)
from jdaviz.core.user_api import LoaderUserApi


@pytest.fixture(scope='function')
def file_drop_resolver(deconfigged_helper):
    """
    Create a FileDropResolver instance with mocked dependencies.
    """
    # Mock these to avoid (GUI) widget creation and whatever comes along with that
    with patch('jdaviz.core.loaders.resolvers.file_drop.file_drop.FileDropMultiple'):
        with patch('jdaviz.core.loaders.resolvers.file_drop.file_drop.reacton') as mock_reacton:
            mock_reacton.render.return_value = (Mock(), Mock())
            # We could also mock the app if needed
            resolver = FileDropResolver(app=deconfigged_helper.app)
            return resolver


class TestFileDropResolverBasic:
    """
    Test FileDropResolver basics.
    """
    def test_init_basic(self, file_drop_resolver):
        """
        Test basic initialization of traitlets and some properties
        in FileDropResolver.
        """
        assert file_drop_resolver._file_info is None
        assert file_drop_resolver.progress == 100
        assert file_drop_resolver.nfiles == 0

        # Test that user_api returns LoaderUserApi instance.
        api = file_drop_resolver.user_api

        # Check that LoaderUserApi was created
        assert isinstance(api, LoaderUserApi)

        assert file_drop_resolver.is_valid is True
        assert file_drop_resolver.default_label is None

        # Test that _on_total_progress updates progress trait.
        for value in [0, 25, 50, 75, 100]:
            file_drop_resolver._on_total_progress(value)
            assert file_drop_resolver.progress == value

    def test_init_creates_widget(self, deconfigged_helper):
        """
        Test that initialization creates file drop widget.
        """
        with (patch('jdaviz.core.loaders.resolvers.file_drop.file_drop.FileDropMultiple')
              as mock_widget):
            with (patch('jdaviz.core.loaders.resolvers.file_drop.file_drop.reacton')
                  as mock_reacton):
                mock_reacton.render.return_value = (Mock(), Mock())

                _ = FileDropResolver(app=deconfigged_helper.app)

                # Check that widget was created with correct parameters
                mock_widget.assert_called_once()
                call_kwargs = mock_widget.call_args[1]
                assert call_kwargs['label'] == 'Drop file here'
                assert call_kwargs['lazy'] is False
                assert 'on_total_progress' in call_kwargs
                assert 'on_file' in call_kwargs

    @pytest.mark.parametrize(('filename', 'result'),
                             [('test_data.csv', 'test_data'),
                              ('my.data.file.fits', 'my.data.file'),
                              ('datafile', 'datafile')])
    def test_default_label_with_file(self, file_drop_resolver, filename, result):
        """
        Test default_label returns file name with different conventions.
        """
        file_drop_resolver._file_info = {'name': filename,
                                         'data': b'some data'}

        assert file_drop_resolver.default_label == result


_FILE_INFO = [{'name': 'file1.csv', 'data': b'data1'},
              {'name': 'file2.csv', 'data': b'data2'},
              {'name': 'file3.csv', 'data': b'data3'}]


class TestFileDropResolverFileHandling:
    """
    Test file upload and handling functionality.
    """
    @pytest.mark.parametrize('file_info', ([_FILE_INFO[0]], _FILE_INFO))
    def test_on_file_updated_single_file(self, file_drop_resolver, file_info):
        """
        Test _on_file_updated with single file.
        """
        # Test that progress updates to 100 after file upload.
        file_drop_resolver.progress = 50
        # Mock the internal methods to track calls but avoid any unintended
        # side effects that we don't need to check here.
        with patch.object(file_drop_resolver, '_resolver_input_updated') as mock_input_updated:
            with patch.object(file_drop_resolver, '_update_format_items') as mock_update_format:
                file_drop_resolver._on_file_updated(file_info)

                assert file_drop_resolver.nfiles == len(file_info)
                assert file_drop_resolver._file_info == file_info[0] if len(file_info) > 1 else file_info  # noqa
                assert file_drop_resolver.progress == 100

                mock_input_updated.assert_called_once()
                mock_update_format.assert_called_once()

    def test_on_file_updated_empty_list(self, file_drop_resolver):
        """
        Test _on_file_updated with empty file list.
        """
        with patch.object(file_drop_resolver, '_resolver_input_updated'):
            with patch.object(file_drop_resolver, '_update_format_items'):
                # This should raise an IndexError
                with pytest.raises(IndexError):
                    file_drop_resolver._on_file_updated([])


class TestFileDropResolverParseInput:
    """
    Test parse_input functionality.
    """
    def test_parse_input_returns_bytesio(self, file_drop_resolver):
        """
        Test that parse_input returns BytesIO object.
        """
        file_drop_resolver._file_info = _FILE_INFO[0]

        result = file_drop_resolver.parse_input()

        assert isinstance(result, io.BytesIO)
        assert result.read() == _FILE_INFO[0]['data']

    def test_parse_input_no_file_info(self, file_drop_resolver):
        """
        Test parse_input when no file is loaded.
        """
        file_drop_resolver._file_info = None
        with pytest.raises(AttributeError):
            file_drop_resolver.parse_input()


def _create_sample_csv_data():
    """
    Create sample CSV data as bytes.
    """
    csv_content = 'col1,col2,col3\n1,2,3\n4,5,6\n7,8,9'
    return csv_content.encode('utf-8')


class TestFileDropResolverParsedInputToTable:
    """
    Test _parsed_input_to_table functionality for various formats.
    """
    def _create_sample_fits_table(self):
        """
        Create sample FITS table data as bytes.
        """
        table = Table({
            'col1': [1, 2, 3],
            'col2': [4, 5, 6],
            'col3': [7, 8, 9]
        })

        # Write to BytesIO using astropy's default write method
        output = io.BytesIO()
        table.write(output, format='fits', overwrite=True)
        output.seek(0)
        return output.getvalue()

    def _create_sample_ascii_data(self):
        """
        Create sample ASCII table data as bytes.
        """
        ascii_content = 'col1 col2 col3\n1 2 3\n4 5 6\n7 8 9'
        return ascii_content.encode('utf-8')

    def test_parsed_input_to_table_csv(self, file_drop_resolver):
        """
        Test parsing CSV format.
        """
        parsed_input = io.BytesIO(_create_sample_csv_data())
        result = file_drop_resolver._parsed_input_to_table(parsed_input)

        assert result is not None
        assert isinstance(result, Table)
        assert len(result) == 3
        assert 'col1' in result.colnames

    def test_parsed_input_to_table_ascii(self, file_drop_resolver):
        """
        Test parsing ASCII format.
        """
        parsed_input = io.BytesIO(self._create_sample_ascii_data())
        result = file_drop_resolver._parsed_input_to_table(parsed_input)

        assert result is not None
        assert isinstance(result, Table)
        assert len(result) == 3

    def test_parsed_input_to_table_fits(self, file_drop_resolver):
        """
        Test parsing FITS format.
        """
        parsed_input = io.BytesIO(self._create_sample_fits_table())
        result = file_drop_resolver._parsed_input_to_table(parsed_input)

        assert result is not None
        assert isinstance(result, Table)
        # Note: _parsed_input_to_table reads primary HDU by default,
        # which is empty. To verify the fixture contains good data,
        # read HDU 1 manually.
        parsed_input.seek(0)
        full_table = Table.read(parsed_input, format='fits', hdu=1)
        assert len(full_table) == 3
        assert 'col1' in full_table.colnames

    def test_parsed_input_to_table_invalid_format(self, file_drop_resolver):
        """
        Test parsing with invalid format.

        Note: astropy.table.Table.read is very lenient and may parse
        unexpected data as a table, so we just verify it doesn't crash.
        """
        invalid_data = b'This is not a valid table format!!@@##'
        parsed_input = io.BytesIO(invalid_data)
        result = file_drop_resolver._parsed_input_to_table(parsed_input)

        assert result is None or isinstance(result, Table)
        if result is not None:
            assert isinstance(result.colnames, list)

    def test_parsed_input_to_table_empty_data(self, file_drop_resolver):
        """
        Test parsing with empty data.

        Note: astropy may create an empty table rather than return None.
        """
        empty_data = b''
        parsed_input = io.BytesIO(empty_data)
        result = file_drop_resolver._parsed_input_to_table(parsed_input)

        # May return None or an empty table
        assert result is None or (isinstance(result, Table) and len(result) == 0)


class TestFileDropResolverEdgeCases:
    """
    Test edge cases.
    """
    def test_file_with_special_conditions(self, file_drop_resolver):
        """
        Test file name with special characters, unicode, and 'large' data.
        """
        file_drop_resolver._file_info = {'name': 'my-file_v2.0 (copy).csv',
                                         'data': b'data'}

        expected = 'my-file_v2.0 (copy)'
        assert file_drop_resolver.default_label == expected

        # Now test with unicode characters
        file_drop_resolver._file_info = {'name': 'és_测试_🌟.csv',
                                         'data': b'data'}
        expected = 'és_测试_🌟'
        assert file_drop_resolver.default_label == expected

        # Simulate a large file by creating 'large' data
        large_data = b'x' * (10 * 1024 * 1024)  # 10 MB
        file_drop_resolver._file_info = {'name': 'large_file.dat',
                                         'data': large_data}
        result = file_drop_resolver.parse_input()

        assert isinstance(result, io.BytesIO)
        # Read in chunks to avoid memory issues in test
        chunk = result.read(1024)
        assert len(chunk) == 1024

    def test_sequential_file_uploads(self, file_drop_resolver):
        """
        Test uploading multiple files sequentially.
        """
        files = [{'name': f'file{i}.csv', 'data': f'data{i}'.encode()}
                 for i in range(3)]

        with patch.object(file_drop_resolver, '_resolver_input_updated'):
            with patch.object(file_drop_resolver, '_update_format_items'):
                for file_info in files:
                    file_drop_resolver.progress = 50
                    file_drop_resolver._on_file_updated([file_info])

                    # Each upload should update _file_info
                    assert file_drop_resolver._file_info['name'] == file_info['name']
                    assert file_drop_resolver.nfiles == 1
                    # Check that progress is reset to 100 after each upload
                    assert file_drop_resolver.progress == 100

    def test_file_info_missing_keys(self, file_drop_resolver):
        """
        Test behavior when file_info is missing name/data keys.
        """
        # Missing 'name' key
        file_drop_resolver._file_info = {'data': b'some data'}

        # default_label should return None
        assert file_drop_resolver.default_label is None

        # Now with missing 'data' key
        file_drop_resolver._file_info = {'name': 'test.csv'}

        # parse_input uses .get() which returns None
        # io.BytesIO(None) creates empty BytesIO, doesn't raise
        result = file_drop_resolver.parse_input()

        assert isinstance(result, io.BytesIO)
        # Should be empty since data was None
        assert result.read() == b''


class TestFileDropResolverIntegration:
    """
    Integration tests combining multiple aspects of functionality.
    """
    def test_full_file_upload_workflow(self, file_drop_resolver):
        """
        Test complete workflow from upload to parsing.
        """
        file_info = {'name': 'data.csv',
                     'data': _create_sample_csv_data()}

        with patch.object(file_drop_resolver, '_resolver_input_updated'):
            with patch.object(file_drop_resolver, '_update_format_items'):
                # Simulate progress updates
                file_drop_resolver._on_total_progress(25)
                assert file_drop_resolver.progress == 25

                file_drop_resolver._on_total_progress(50)
                assert file_drop_resolver.progress == 50

                # Upload file
                file_drop_resolver._on_file_updated([file_info])
                assert file_drop_resolver.nfiles == 1
                assert file_drop_resolver.progress == 100

                # Check default label
                assert file_drop_resolver.default_label == 'data'

                # Parse input
                parsed = file_drop_resolver.parse_input()
                assert isinstance(parsed, io.BytesIO)

                # Convert to table
                table = file_drop_resolver._parsed_input_to_table(parsed)
                assert table is not None
                assert isinstance(table, Table)

    def test_error_recovery_on_invalid_file(self, file_drop_resolver):
        """
        Test that resolver handles invalid files gracefully.
        """
        invalid_file = {'name': 'invalid.xyz',
                        'data': b'random binary data \x00\x01\x02'}

        with patch.object(file_drop_resolver, '_resolver_input_updated'):
            with patch.object(file_drop_resolver, '_update_format_items'):
                file_drop_resolver._on_file_updated([invalid_file])

                # Should still update state
                assert file_drop_resolver.nfiles == 1
                assert file_drop_resolver.default_label == 'invalid'

                # Parsing - astropy is lenient, may parse as table
                parsed = file_drop_resolver.parse_input()
                table = file_drop_resolver._parsed_input_to_table(parsed)
                # Either returns None or a table (astropy is very lenient)
                assert table is None or isinstance(table, Table)



import io
import os
import re
import pytest
import numpy as np
import astropy.units as u
from astropy.table import Table
from specutils import SpectralRegion

from jdaviz.core.template_mixin import TableMixin


def test_spectralsubsetselect(specviz_helper, spectrum1d):
    # apply mask to spectrum to check selected subset is masked:
    mask = spectrum1d.flux < spectrum1d.flux.mean()
    spectrum1d.mask = mask

    specviz_helper.load_data(spectrum1d)
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    # create a "Subset 1" entry
    subset_plugin = specviz_helper.plugins['Subset Tools']
    subset_plugin.import_region(SpectralRegion(6500 * spectrum1d.spectral_axis.unit,
                                               7400 * spectrum1d.spectral_axis.unit))

    # model fitting uses the mixin
    p = specviz_helper.app.get_tray_item_from_name('g-model-fitting')
    assert len(p.spectral_subset.labels) == 2  # Entire Spectrum, Subset 1
    assert len(p.spectral_subset_items) == 2
    assert p.spectral_subset_selected == 'Entire Spectrum'
    assert p.spectral_subset_selected_has_subregions is False
    assert p.spectral_subset.selected_obj is None
    p.spectral_subset_selected = 'Subset 1'
    assert p.spectral_subset_selected_has_subregions is False
    assert p.spectral_subset.selected_obj is not None
    expected_min = spectrum1d.spectral_axis[spectrum1d.spectral_axis.value >= 6500][0]
    expected_max = spectrum1d.spectral_axis[spectrum1d.spectral_axis.value <= 7400][-1]
    np.testing.assert_allclose(expected_min.value, 6666.66666667, atol=1e-5)
    np.testing.assert_allclose(expected_max.value, 7333.33333333, atol=1e-5)
    assert p.spectral_subset.selected_min_max(spectrum1d) == (6500 * u.AA, 7400 * u.AA)

    # check selected subset mask available via API:
    expected_mask_with_spectral_subset = (
        (spectrum1d.wavelength.to_value(u.AA) < 6500) |
        (spectrum1d.wavelength.to_value(u.AA) > 7400)
    )
    assert np.all(
        expected_mask_with_spectral_subset == p.spectral_subset.selected_subset_mask
    )

    assert p.spectral_subset.app == p.app
    assert p.spectral_subset.spectrum_viewer == sv

    # line analysis uses custom components, one of which is still named spectral_subset
    p = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert len(p.spectral_subset.labels) == 2  # Entire Spectrum, Subset 1
    assert len(p.spectral_subset_items) == 2
    assert p.spectral_subset_selected == 'Entire Spectrum'
    assert p.spectral_subset.selected_obj is None
    p.spectral_subset_selected = 'Subset 1'
    assert p.spectral_subset.selected_obj is not None


@pytest.mark.filterwarnings('ignore:No observer defined on WCS')
def test_viewer_select(cubeviz_helper, spectrum1d_cube):
    app = cubeviz_helper.app
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    fv = app.get_viewer("flux-viewer")
    sv = app.get_viewer('spectrum-viewer')

    # export plugin uses the mixin
    p = cubeviz_helper.plugins['Export']
    # 2 viewers available (no uncertainty cube loaded)
    assert len(p.viewer.ids) == 2
    assert len(p.viewer.references) == 2
    assert len(p.viewer.labels) == 2
    assert p.viewer.selected_obj == fv

    # set by reference
    p.viewer = 'spectrum-viewer'
    assert p.viewer.selected_obj == sv

    # try setting based on id instead of reference
    p.viewer = p.viewer.ids[0]
    assert p.viewer.selected == p.viewer.labels[0]


class TestObserveTraitletsForRelevancy:

    fake_traitlets = ['fake_traitlet1', 'fake_traitlet2']

    set_relevant_msg = 'stale message'
    new_set_relevant_msg = 'fresh message'
    sad_msg = "I don't want to be made irrelevant :("

    def if_fake_truthy(self, *args):
        self.set_relevant_msg = self.new_set_relevant_msg

    def setup_plugin_obj_and_traitlets(self, config_helper):
        plugins = config_helper.plugins

        plugin_obj = plugins['Test Fake Plugin']._obj

        self.if_all_truthy = plugin_obj.relevant_if_all_truthy
        self.if_any_truthy = plugin_obj.relevant_if_any_truthy

        return plugin_obj, [trait_name for trait_name in plugin_obj.traits()
                            if getattr(plugin_obj, trait_name, False)]

    def test_setup_relevant_if_truthy(self, deconfigged_helper):
        deconfigged_plugin_obj, traitlets = self.setup_plugin_obj_and_traitlets(
            deconfigged_helper)

        default_irrelevant_msg = (f"At least one of or all of {', '.join(traitlets)} "
                                  f"are not available")

        # Covers no _traitlets_to_observe set
        with pytest.raises(AttributeError,
                           match=re.escape('_traitlets_to_observe has not been '
                                           'set yet (by `observe_traitlets_for_relevancy`).')):
            deconfigged_plugin_obj._setup_relevant_if_truthy(None)

        # Covers _traitlets_to_observe set and default message
        deconfigged_plugin_obj._traitlets_to_observe = traitlets
        traitlets_from_setup, truthy_traitlets, irrelevant_msg = (
            deconfigged_plugin_obj._setup_relevant_if_truthy(None))
        assert traitlets_from_setup == traitlets
        assert truthy_traitlets == traitlets
        assert irrelevant_msg == default_irrelevant_msg

        # Covers traitlets passed in as arg and default message
        traitlets_from_setup, truthy_traitlets, irrelevant_msg = (
            deconfigged_plugin_obj._setup_relevant_if_truthy(traitlets))
        assert traitlets_from_setup == traitlets
        assert truthy_traitlets == traitlets
        assert irrelevant_msg == default_irrelevant_msg

    @pytest.mark.parametrize('method_type', ['all', 'any'])
    def test_relevant_if_truthy(self, deconfigged_helper, method_type):
        deconfigged_plugin_obj, traitlets = self.setup_plugin_obj_and_traitlets(
            deconfigged_helper)

        method = getattr(self, f'if_{method_type}_truthy')

        # Checking the default argument (None) when nothing is observed yet
        with pytest.raises(AttributeError,
                           match=re.escape('_traitlets_to_observe has not been set yet '
                                           '(by `observe_traitlets_for_relevancy`).')):
            method()

        # Check that irrelevant message is set correctly for relevant traitlets
        result = method(traitlets)
        assert result == ''

        # Check that irrelevant message is set correctly for solely irrelevant traitlets
        result = method(self.fake_traitlets)
        assert result == (f"At least one of or all of {', '.join(self.fake_traitlets)} "
                          f"are not available")

        # Check that irrelevant message is set correctly for relevant + irrelevant traitlets
        traitlets += self.fake_traitlets
        result_default_msg = method(traitlets)

        # Check that irrelevant message is set correctly for relevant + irrelevant traitlets
        if method_type == 'all':
            assert result_default_msg == (f"At least one of or all of {', '.join(traitlets)} "
                                          f"are not available")
        elif method_type == 'any':
            assert result_default_msg == ''

    @pytest.mark.parametrize('method_type', ['all', 'any', 'fake'])
    def test_default_set_relevant(self, deconfigged_helper, method_type):
        deconfigged_plugin_obj, traitlets = self.setup_plugin_obj_and_traitlets(deconfigged_helper)
        deconfigged_plugin_obj._irrelevant_msg_callback = getattr(self, f'if_{method_type}_truthy')
        deconfigged_plugin_obj._traitlets_to_observe = traitlets

        # Real traitlets
        deconfigged_plugin_obj._set_relevant()
        if deconfigged_plugin_obj.irrelevant_msg is None:
            # i.e. when the fake method is used
            assert self.set_relevant_msg == self.new_set_relevant_msg
            self.set_relevant_msg = 'newly staled'
        else:
            assert deconfigged_plugin_obj.irrelevant_msg == ''

        # Real + Fake Traitlets
        # Thanks to mutability we shouldn't also have to update '_traitlets_to_observe'
        traitlets += self.fake_traitlets
        deconfigged_plugin_obj._set_relevant()
        if deconfigged_plugin_obj.irrelevant_msg is None:
            # i.e. when the fake method is used
            assert self.set_relevant_msg == self.new_set_relevant_msg
            self.set_relevant_msg = 'newly staled'

        else:
            if method_type == 'all':
                assert deconfigged_plugin_obj.irrelevant_msg == (f"At least one of or all of "
                                                                 f"{', '.join(traitlets)} "
                                                                 f"are not available")
            elif method_type == 'any':
                assert deconfigged_plugin_obj.irrelevant_msg == ''

    @pytest.mark.parametrize('additional_traitlets', [[], fake_traitlets])
    def test_observe_traitlets_for_relevancy(self,
                                             deconfigged_helper, additional_traitlets):
        deconfigged_plugin_obj, traitlets = self.setup_plugin_obj_and_traitlets(deconfigged_helper)
        traitlets += additional_traitlets

        if len(additional_traitlets) == 0:
            # i.e. irrelevant_msg is an empty string
            expected_msg = ''
        else:
            # Real + fake so the default (if_all_truthy) is the observer
            expected_msg = f"At least one of or all of {', '.join(traitlets)} are not available"

        # Testing default functionality
        deconfigged_plugin_obj.observe_traitlets_for_relevancy(traitlets)
        assert deconfigged_plugin_obj._traitlets_to_observe == traitlets
        assert (deconfigged_plugin_obj._irrelevant_msg_callback() ==
                deconfigged_plugin_obj.relevant_if_all_truthy())

        assert deconfigged_plugin_obj.irrelevant_msg == expected_msg

        # Now to check that the traitlets are observed as expected
        # and that _set_relevant() therefore runs as expected
        observed_traitlets = deconfigged_plugin_obj._trait_notifiers
        # 'Observed' as expected
        assert set(traitlets).issubset(set(observed_traitlets.keys()))

        # Keep trait_name around for future debugging if necessary
        observed_traitlet_methods = {trait_name: observed_traitlets[trait_name]['change']
                                     for trait_name in traitlets}

        # Check to see if _set_relevant is available and runs
        # when accessing the observed quantity
        for _, observe_methods in observed_traitlet_methods.items():
            deconfigged_plugin_obj.irrelevant_msg = self.sad_msg

            # Run the method to update irrelevant_msg
            # traits may be observed by more than one method
            # traitlet ObserverHandlers don't have __name__ so we use getattr
            all_results = [method() for method in observe_methods
                           if getattr(method, '__name__', None) == '_set_relevant']

            # _set_relevant() ran as expected
            assert deconfigged_plugin_obj.irrelevant_msg == expected_msg
            # And double check that the function output what we expect (Nones in this case)
            assert deconfigged_plugin_obj._set_relevant() in all_results

    @pytest.mark.parametrize('additional_traitlets', [[], fake_traitlets])
    def test_observe_traitlets_for_relevancy_with_fake_callback(
            self, deconfigged_helper, additional_traitlets):
        deconfigged_plugin_obj, traitlets = self.setup_plugin_obj_and_traitlets(
            deconfigged_helper)
        traitlets += additional_traitlets

        deconfigged_plugin_obj.observe_traitlets_for_relevancy(
            traitlets, irrelevant_msg_callback=self.if_fake_truthy)
        assert deconfigged_plugin_obj._traitlets_to_observe == traitlets
        assert (deconfigged_plugin_obj._irrelevant_msg_callback() ==
                self.if_fake_truthy())

        # Checking that _set_relevant ran and resetting the message
        assert self.set_relevant_msg == self.new_set_relevant_msg

        # Now to check that the traitlets are observed as expected
        # and that _set_relevant() therefore runs as expected
        observed_traitlets = deconfigged_plugin_obj._trait_notifiers
        # 'Observed' as expected
        assert set(traitlets).issubset(set(observed_traitlets.keys()))

        # Keep trait_name around for future debugging if necessary
        observed_traitlet_methods = {trait_name: observed_traitlets[trait_name]['change']
                                     for trait_name in traitlets}

        # Check to see if _set_relevant is available and runs
        # when accessing the observed quantity
        for _, observe_methods in observed_traitlet_methods.items():
            not_fresh_msg = 'less stale message but still not great'
            self.set_relevant_msg = not_fresh_msg

            # Run the method to update irrelevant_msg
            # traits may be observed by more than one method
            # traitlet ObserverHandlers don't have __name__ so we use getattr
            _ = [method() for method in observe_methods
                 if getattr(method, '__name__', None) == '_set_relevant']

            # Check that self.set_relevant_msg is overwritten by our if_fake_truthy
            # behind the scenes
            assert self.set_relevant_msg != not_fresh_msg


def test_select_plugin_component_map_value(imviz_helper, multi_extension_image_hdu_wcs):
    ldr = imviz_helper.loaders['object']
    ldr.object = multi_extension_image_hdu_wcs
    selection_obj = ldr.importer.extension

    # Check default selection
    assert selection_obj.selected == [selection_obj.choices[0]]

    # Check set selected directly via wildcard
    selection_obj.selected = '*'
    assert selection_obj.selected == selection_obj.choices
    choices_before = selection_obj.choices

    # Check set selected directly via ? wildcard
    selection_obj.selected = '*MASK,?]'
    assert selection_obj.selected == [selection_obj.choices[1]]
    choices_before = selection_obj.choices

    selection_obj.fake_attribute = 'item1'
    # Check that a different attribute *does not* trigger the wildcard logic
    # i.e. it gets set as-is and doesn't go through `choices`
    selection_obj.fake_attribute = ['fake * item1', 'fake item2']
    assert selection_obj.fake_attribute == ['fake * item1', 'fake item2']
    assert selection_obj.choices == choices_before


class FakeTable(TableMixin):
    template = ''

    def __init__(self, session, catalog, *args, **kwargs):
        self.session = session
        self._plugin_name = 'test-fake-table'
        super().__init__(*args, **kwargs)
        self.table._qtable = catalog


def astropy_table_write_formats():
    table_obj = Table({'a': [1, 2], 'b': [3, 4]})
    buf = io.StringIO()
    table_obj.write.list_formats(out=buf)
    output = buf.getvalue()

    # Output looks like this:
    #            Format           Read Write Auto-identify Deprecated
    # --------------------------- ---- ----- ------------- ----------
    #                       ascii  Yes   Yes            No
    #                ascii.aastex  Yes   Yes            No
    #                 ascii.basic  Yes   Yes            No

    valid_formats = []
    for line in output.splitlines():
        if line.strip().startswith('Format') or line.strip().startswith('--------'):
            continue

        split_line = line.strip().split()
        if len(split_line) == 5 and split_line[-1] == 'Yes':
            # Deprecated
            continue
        valid_formats.append(line.strip().split()[0])

    return valid_formats


def test_export_table_special_cases():
    valid_formats = astropy_table_write_formats()

    # These are special cases for formats that are currently supported by astropy but require
    # some specific details to be handled. If these get removed in the future, then
    # we can remove the special handling in `export_table`.
    assert {'asdf',
            'hdf5',
            'ascii.tdat',
            'votable',
            'parquet',
            'parquet.votable',
            'votable.parquet',
            'ascii.ipac'}.issubset(set(valid_formats))


@pytest.mark.parametrize('valid_format', astropy_table_write_formats())
def test_export_table(deconfigged_helper, sky_coord_only_source_catalog,
                      tmp_path, valid_format):

    table_obj = FakeTable(deconfigged_helper.app.session,
                          sky_coord_only_source_catalog)

    tmp_filename = tmp_path / 'temp_table'
    filename = f"{tmp_filename}_{valid_format}"

    if 'parquet' in valid_format:
        # Check for partial match for parquet and votable.parquet
        try:
            table_obj.export_table(filename,
                                   format=valid_format,
                                   overwrite=True)
            assert os.path.isfile(filename)
        except ModuleNotFoundError as me:
            assert 'This is not a default dependency of jdaviz.' in str(me)

    elif 'asdf' in valid_format:
        # Check asdf
        table_obj.export_table(filename, format=valid_format)
        assert os.path.isfile(filename)
        with pytest.raises(FileExistsError, match=r'exists and overwrite=False'):
            table_obj.export_table(filename,
                                   format=valid_format,
                                   overwrite=False)

    else:
        table_obj.export_table(filename, format=valid_format)
        assert os.path.isfile(filename)



import os
from traitlets import Any, Bool, List, Unicode, observe
from glue.core.message import (DataCollectionAddMessage,
                               DataCollectionDeleteMessage)

from jdaviz.core.events import NewViewerMessage, SnackbarMessage
from jdaviz.core.registries import viewer_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        AutoTextField,
                                        ViewerSelectCreateNew,
                                        with_spinner)
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.utils import (standardize_metadata,
                          _wcs_only_label,
                          CONFIGS_WITH_LOADERS,
                          create_data_hash)

__all__ = ['BaseImporter', 'BaseImporterToDataCollection', 'BaseImporterToPlugin']


class BaseImporter(PluginTemplateMixin):
    # preference order of parsers, by registry name.  If empty, the first found match will
    # be used by default.  If not empty, the first match in the list will be used (including
    # over any parsers not included in the list).  If not empty but no valid parsers are in
    # the list, the first remaining match will be used.
    parser_preference = []

    import_disabled = Bool(False).tag(sync=True)
    import_spinner = Bool(False).tag(sync=True)

    existing_data_in_dc = List([]).tag(sync=True)

    def __init__(self, app, resolver, parser, input, **kwargs):
        self._input = input
        self._parser = parser
        self._resolver = resolver
        super().__init__(app, **kwargs)

        # Doing this in app instead of here avoids a lot of unnecessary overhead
        # from all the importers in memory
        self.app.observe(self._update_existing_data_in_dc_traitlet, 'existing_data_in_dc')
        self._update_existing_data_in_dc_traitlet()

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def _apply_kwargs(self, kwargs):
        user_api = self.user_api
        applied_kwargs = []
        for k, v in kwargs.items():
            if hasattr(user_api, k) and v is not None:
                setattr(user_api, k, v)
                applied_kwargs.append(k)

        return applied_kwargs

    @property
    def is_valid(self):
        # override by subclass
        return False  # pragma: nocover

    @property
    def resolver(self):
        return self._resolver

    @property
    def input(self):
        return self._input

    @property
    def output(self):
        # override by subclass
        return self.input

    def _update_existing_data_in_dc_traitlet(self, change={}):
        self.existing_data_in_dc = self.app.existing_data_in_dc

    def reset_and_check_existing_data_in_dc(self, change={}):
        """
        Check if the data to be imported appears to already exist in the data collection
        based on the data hash.  If so, update the existing_data_in_dc traitlet
        accordingly and display a warning snackbar message.
        """
        if not hasattr(self, 'data_hashes'):
            # If we do this here instead of at init, then we shouldn't get errors
            # from attempting to access unavailable importer attributes from 'output'
            self.data_hashes = [create_data_hash(self.output)]

        if not hasattr(self, 'hash_map_to_label'):
            self.hash_map_to_label = {dh: '' for dh in self.data_hashes}

        dc_labels = []
        loader_labels = []
        existing_data_in_dc = [(data.meta.get('_data_hash'),
                                data.label,
                                self.hash_map_to_label[data.meta.get('_data_hash')])
                               for data in self.app.data_collection
                               if data.meta.get('_data_hash') in self.data_hashes]

        if len(existing_data_in_dc) > 0:
            existing_data_in_dc, dc_labels, loader_labels = zip(*existing_data_in_dc)
        self.app.existing_data_in_dc = list(existing_data_in_dc)

        # Only need to display the message once
        if len(dc_labels) > 0:
            if any(self.hash_map_to_label.values()):
                msg = 'Selected data appears to be identical to existing data.\n'
                for dc_label, loader_label in zip(dc_labels, loader_labels):
                    msg += f"{loader_label} <=> {dc_label}\n"
            else:
                msg = (f"Selected data appears to be identical "
                       f"to existing data ({', '.join(dc_labels)}).")

            self.app.hub.broadcast(SnackbarMessage(msg.rstrip('\n'), sender=self, color='warning'))
            # TODO: Allow for now but implement a disabled message near the import button
            #  or indicate that the import will be a re-import And if allowing re-import,
            #  there's a bug that needs to be squashed... (second import doesn't show in viewer)
            #  but remains in app.
            # self.data_label_invalid_msg = msg

    @property
    def target(self):
        raise NotImplementedError("Importer subclass must implement target")  # pragma: nocover

    def __call__(self):
        # override by subclass - should act on self.output and load into jdaviz
        raise NotImplementedError("Importer subclass must implement __call__")  # pragma: nocover

    @property
    def user_api(self):
        return ImporterUserApi(self)

    def vue_import_clicked(self, *args, **kwargs):
        self._resolver.load()


class BaseImporterToDataCollection(BaseImporter):
    data_label_value = Unicode().tag(sync=True)
    data_label_default = Unicode().tag(sync=True)
    data_label_auto = Bool(True).tag(sync=True)
    data_label_invalid_msg = Unicode().tag(sync=True)

    viewer_create_new_items = List([]).tag(sync=True)
    viewer_create_new_selected = Unicode().tag(sync=True)

    viewer_items = List([]).tag(sync=True)
    viewer_selected = Any([]).tag(sync=True)
    viewer_multiselect = Bool(True).tag(sync=True)

    viewer_label_value = Unicode().tag(sync=True)
    viewer_label_default = Unicode().tag(sync=True)
    viewer_label_auto = Bool(True).tag(sync=True)
    viewer_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, app, resolver, parser, input, **kwargs):
        super().__init__(app, resolver, parser, input, **kwargs)
        self.data_label = AutoTextField(self, 'data_label_value',
                                        'data_label_default',
                                        'data_label_auto',
                                        'data_label_invalid_msg')

        self.data_label_default = self.app.return_unique_name(self._registry_label)

        self.viewer = ViewerSelectCreateNew(self, 'viewer_items',
                                            'viewer_selected',
                                            'viewer_create_new_items',
                                            'viewer_create_new_selected',
                                            'viewer_label_value',
                                            'viewer_label_default',
                                            'viewer_label_auto',
                                            'viewer_label_invalid_msg',
                                            multiselect='viewer_multiselect',
                                            default_mode='empty')

        self.hub.subscribe(self, DataCollectionAddMessage,
                           handler=lambda _: self._on_label_changed())
        self.hub.subscribe(self, DataCollectionDeleteMessage,
                           handler=lambda _: self._on_label_changed())
        self._on_label_changed()

        supported_viewers = self._get_supported_viewers()
        if self.app.config in ('deconfigged', 'imviz', 'lcviz', 'rampviz'):
            if self.app.config == 'imviz':
                # only allow image viewers
                supported_viewers = [item for item in supported_viewers
                                     if item.get('reference') == 'imviz-image-viewer']
            self.viewer_create_new_items = [item for item in supported_viewers
                                            if item.pop('allow_create', True)]

        # for now, we'll use the same list of viewers that can be created
        # as the filter on existing viewers.  Eventually we may want this
        # to be more flexible so that custom viewers can show up in this
        # list (or both lists) without having to modify the importer - perhaps
        # by both viewers and importers describing their supported data shapes
        # and physical types
        def viewer_in_registry_names(viewer):
            classes = [viewer_registry.members.get(item.get('reference')).get('cls')
                       for item in supported_viewers]
            return isinstance(viewer, tuple(classes))
        self.viewer.add_filter(viewer_in_registry_names)
        self.viewer.select_default()

    @staticmethod
    def _get_supported_viewers():
        raise NotImplementedError("Importer subclass must implement _get_supported_viewers")  # noqa pragma: nocover

    @property
    def ignore_viewers_with_cls(self):
        return ()

    @property
    def default_data_label_from_resolver(self):
        if self._resolver.parsed_input_is_query and self._resolver.treat_table_as_query:
            url = self._resolver.get_selected_url()
            path = os.path.splitext(os.path.basename(url.strip()))[0]
            if "product_name=" in path:
                path = path.split("product_name=")[1]
            return path
        return self._resolver.default_label

    @property
    def target(self):
        if len(self.viewer.create_new.choices) > 0:
            return {'type': 'viewer',
                    'icon': 'mdi-window-maximize',
                    'label': self.viewer.create_new.choices[0]}
        else:
            return {}

    @observe('data_label_value')
    def _on_label_changed(self, msg={}):
        if not len(self.data_label_value.strip()):
            # strip will raise the same error for a label of all spaces
            self.data_label_invalid_msg = 'data_label must be provided'
            return

        # ensure the default label is unique for the data-collection
        self.data_label_default = self.app.return_unique_name(self.data_label_default)

        for data in self.app.data_collection:
            if self.data_label_value == data.label:
                self.data_label_invalid_msg = f"data_label='{self.data_label_value}' already in use"
                return

        self.data_label_invalid_msg = ''

    @observe('data_label_invalid_msg', 'viewer_label_invalid_msg')
    def _set_import_disabled(self, change={}):
        self.import_disabled = (len(self.data_label_invalid_msg) > 0
                                or len(self.viewer_label_invalid_msg) > 0)

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return physical_type

    def add_to_data_collection(self, data, data_label=None, data_hash=None,
                               parent=None,
                               viewer_select=None,
                               cls=None):

        if data_label is None:
            data_label = self.data_label_value.strip()
        else:
            data_label = data_label.strip()

        if hasattr(data, 'meta'):
            try:
                data.meta = standardize_metadata(data.meta)
            except TypeError:
                pass

        # store the original input class so that get_data can default to the
        # same class as the input
        cls = cls if cls is not None else data.__class__
        if not hasattr(data, 'meta'):
            data.meta = {}
        if not isinstance(data.meta, dict):
            # ensure that data.meta is a dictionary and supports item assignment
            data.meta = dict(data.meta)
        data.meta['_native_data_cls'] = cls
        data.meta['_importer'] = self.__class__.__name__

        # Create a hashed representation of the data if not already present
        data.meta['_data_hash'] = data_hash if data_hash is not None else create_data_hash(data)

        self.app.add_data(data, data_label=data_label)
        if parent is not None:
            self.app._set_assoc_data_as_child(data_label, parent)

        def _physical_type_from_component(comp_id, comp):
            import astropy.units as u
            try:
                comp_units = comp.units
                if comp_units is None or comp_units == '':
                    return comp_units, None
                return comp_units, str(u.Unit(comp_units).physical_type)
            except (ValueError, TypeError, AttributeError):
                return comp_units, None

        new_dc_entry = self.app.data_collection[data_label]
        for comp_id in new_dc_entry.components:
            comp_units, physical_type = _physical_type_from_component(str(comp_id),
                                                                      new_dc_entry.get_component(comp_id))  # noqa
            comp_id._component_type = self.assign_component_type(str(comp_id),
                                                                 new_dc_entry.get_component(comp_id),  # noqa
                                                                 comp_units, physical_type)

        if self.app.config in CONFIGS_WITH_LOADERS:
            self.app._link_new_data_by_component_type(data_label)

        viewer_select = viewer_select if viewer_select is not None else self.viewer
        if viewer_select.create_new.selected:
            viewer_reference = viewer_select.create_new.selected_item.get('reference')
            viewer_label = viewer_select.new_label.value.strip()

            viewer_dict = viewer_registry.members.get(viewer_reference)
            viewer_cls = viewer_dict.get('cls')
            self.app._on_new_viewer(NewViewerMessage(viewer_cls, data=None, sender=self.app),
                                    vid=viewer_label,
                                    name=viewer_label,
                                    open_data_menu_if_empty=False)
            viewer = self.app._jdaviz_helper.viewers.get(viewer_label)
            viewer.data_menu.add_data(data_label)

            # default to selecting this new viewer for next import
            viewer_select.create_new.selected = ''
            viewer_select.selected = [viewer_label]

        elif len(viewer_select.selected) == 0:
            # just send a snackbar as feedback
            if len(self.app._jdaviz_helper.viewers):
                msg = f"{data_label} loaded without any viewers selected - add manually from viewer data-menu"  # noqa
            else:
                msg = f"{data_label} loaded but no viewers were created.  Create viewers manually and add data from data-menu"  # noqa
            if not new_dc_entry.meta.get(_wcs_only_label, False):
                self.app.hub.broadcast(SnackbarMessage(msg, sender=self, color='warning'))
        else:
            failed_viewers = []
            exceptions = []
            for viewer_label in viewer_select.selected:
                viewer = self.app._jdaviz_helper.viewers.get(viewer_label)
                try:
                    viewer.data_menu.add_data(data_label)
                except Exception as e:
                    failed_viewers.append(viewer_label)
                    exceptions.append(str(e))
            if len(failed_viewers) > 0:
                msg = f"Failed to add {data_label} to viewers: {', '.join(failed_viewers)}"
                self.app.hub.broadcast(SnackbarMessage(msg, sender=self, color='error',
                                                       traceback=exceptions))

    @with_spinner('import_spinner')
    def __call__(self):
        if self.data_label_invalid_msg:
            raise ValueError(self.data_label_invalid_msg)
        if self.viewer.create_new.selected != '' and self.viewer_label_invalid_msg:
            raise ValueError(self.viewer_label_invalid_msg)
        # NOTE: if data hashing performance becomes an issue for importers that
        # don't overwrite __call__, we can pass the pre-computed hash from
        # self.data_hashes as a kwarg here
        self.add_to_data_collection(self.output)


class BaseImporterToPlugin(BaseImporter):
    @property
    def default_plugin(self):
        raise NotImplementedError("Importer subclass must implement default_plugin")  # noqa pragma: nocover

    @property
    def target(self):
        return {'type': 'plugin',
                'icon': 'mdi-toy-brick-outline',
                'label': self.default_plugin}

    @property
    def has_default_plugin(self):
        return self.default_plugin in self.app._jdaviz_helper.plugins



import numpy as np

from asdf import AsdfFile
from astropy import units as u
from astropy.io import fits
from astropy.nddata import StdDevUncertainty
from astropy.wcs import WCS
from functools import cached_property
from glue.core import HubListener
from ipyvuetify import VuetifyTemplate
from specutils import Spectrum
from traitlets import List, Unicode, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.template_mixin import SelectFileExtensionComponent
from jdaviz.core.unit_conversion_utils import check_if_unit_is_per_solid_angle
from jdaviz.core.custom_units_and_equivs import PIX2, _eqv_flux_to_sb_pixel
from jdaviz.utils import (standardize_metadata,
                          PRIHDR_KEY,
                          SPECTRAL_AXIS_COMP_LABELS,
                          _get_celestial_wcs)

__all__ = ['SpectrumInputExtensionsMixin', '_spectrum_assign_component_type']


def _spectrum_assign_component_type(comp_id, comp, units, physical_type):
    if not len(units) and comp_id == 'flux':
        units = 'ct'
    if units in ('ct', 'pixel'):
        physical_type = units

    if physical_type is None:
        return None
    if comp_id in SPECTRAL_AXIS_COMP_LABELS:
        if physical_type in ('frequency', 'length', 'pixel'):
            # link frequency to wavelength
            return 'spectral_axis'
        return f'spectral_axis:{physical_type}'
    if comp_id == 'uncertainty':
        # don't link with flux columns
        return f'uncertainty:{physical_type}'
    return physical_type


class SpectrumInputExtensionsMixin(VuetifyTemplate, HubListener):
    input_type = Unicode().tag(sync=True)

    extension_items = List().tag(sync=True)
    extension_selected = Unicode().tag(sync=True)

    unc_extension_items = List().tag(sync=True)
    unc_extension_selected = Unicode().tag(sync=True)

    mask_extension_items = List().tag(sync=True)
    mask_extension_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.input, fits.HDUList):
            self.input_type = 'fits:hdulist'
            ext_options = [{'label': f"{index}: {hdu.name}",
                            'name': hdu.name,
                            'ver': hdu.ver,
                            'name_ver': f"{hdu.name},{hdu.ver}",
                            'index': index,
                            'obj': hdu}
                           for index, hdu in enumerate(self.input)
                           ]

        elif isinstance(self.input, AsdfFile) and 'roman' in self.input:
            self.input_type = 'asdf:roman'
            ext_options = [{'label': f"{ind}: {k}",
                            'name': k,
                            'name_ver': k,
                            'index': ind,
                            'obj': ext}
                           for ind, (k, ext) in enumerate(self.input['roman']['data'].items())
                           ]

        elif isinstance(self.input, Spectrum):
            self.input_type = 'specutils:spectrum'
            ext_options = [{'label': f'spectrum.{attr}',
                            'name': attr,
                            'name_ver': None,
                            'index': ind+1,  # to match indexing of HDUList for load_data defaults
                            'obj': self.input}
                           for ind, attr in enumerate(('flux', 'uncertainty', 'mask'))
                           if getattr(self.input, attr, None) is not None
                           ]
        else:
            raise TypeError("Input type not supported for SpectrumInputExtensionsMixin")

        ext_options += [{'label': 'None',
                         'name': '',
                         'name_ver': '',
                         'index': len(ext_options),
                         'obj': None}]

        self.extension = SelectFileExtensionComponent(self,
                                                      items='extension_items',
                                                      selected='extension_selected',
                                                      manual_options=ext_options,
                                                      filters=[self.is_valid_flux])
        self.unc_extension = SelectFileExtensionComponent(self,
                                                          items='unc_extension_items',
                                                          selected='unc_extension_selected',
                                                          manual_options=ext_options,
                                                          filters=[self.is_valid_unc])
        self.mask_extension = SelectFileExtensionComponent(self,
                                                           items='mask_extension_items',
                                                           selected='mask_extension_selected',
                                                           manual_options=ext_options,
                                                           filters=[self.is_valid_mask])

    def _cleanup(self):
        for attr in ('extension', 'unc_extension', 'mask_extension'):
            if not hasattr(self, attr):
                continue

            for ext in getattr(self, attr).manual_options:
                try:
                    del ext['obj'].data
                except Exception:  # nosec
                    pass

        if self.input_type == 'fits:hdulist':
            for hdu in self.input:
                try:
                    del hdu.data
                except Exception:  # nosec
                    pass

        self._clear_cache('spectrum')

    @property
    def supported_flux_ndim(self):
        return 2

    @property
    def default_spectral_axis_index(self):
        return 1

    def is_valid_flux(self, item):
        if self.input_type == 'fits:hdulist':
            return self.hdu_is_valid_flux(item)
        if self.input_type == 'asdf:roman':
            return self.asdf_roman_is_valid_flux(item)
        if self.input_type == 'specutils:spectrum':
            return item.get('name') == 'flux'
        raise NotImplementedError("Unknown input type")

    def hdu_is_valid_flux(self, item):
        """
        Check if the HDU is valid to be imported for the flux in a Spectrum.

        Parameters
        ----------
        hdu : `astropy.io.fits.hdu.base.HDUBase`
            The HDU to check.

        Returns
        -------
        bool
            True if the HDU is a valid light curve HDU, False otherwise.
        """
        if item.get('label') == 'None':
            # require selection for flux
            return False
        hdu = item.get('obj')
        return (len(getattr(hdu, 'shape', [])) == self.supported_flux_ndim
                and ('DISPAXIS' in hdu.header
                     or hdu.header.get('CTYPE1', '') == 'WAVE'
                     or hdu.header.get('EXTNAME', '') == 'FLUX'))

    def asdf_roman_is_valid_flux(self, item):
        if item.get('label') == 'None':
            # require selection for flux
            return False
        if self.supported_flux_ndim == 1:
            return all(k in item.get('obj').keys() for k in ["wl", "flux"])
        elif self.supported_flux_ndim == 2:
            return all(k in item.get('obj').keys() for k in ["spectrum", "wavelength"])
        else:
            raise NotImplementedError("Only 1D and 2D spectra are supported for ASDF Roman data")

    def is_valid_unc(self, item):
        if item.get('label') == 'None':
            return True
        if self.input_type == 'fits:hdulist':
            return self.hdu_is_valid_unc(item)
        if self.input_type == 'asdf:roman':
            return self.asdf_roman_is_valid_unc(item)
        if self.input_type == 'specutils:spectrum':
            return item.get('name') == 'uncertainty'
        raise NotImplementedError("Unknown input type")

    def hdu_is_valid_unc(self, item):
        """
        Check if the HDU is valid to be imported for the uncertainty in a Spectrum.

        Parameters
        ----------
        hdu : `astropy.io.fits.hdu.base.HDUBase`
            The HDU to check.

        Returns
        -------
        bool
            True if the HDU is a valid light curve HDU, False otherwise.
        """
        hdu = item.get('obj')
        return (len(getattr(hdu, 'shape', [])) == self.supported_flux_ndim
                and hdu.header.get('EXTNAME', '') == 'ERR')

    def asdf_roman_is_valid_unc(self, item):
        # TODO: should this instead have options just to include or not
        # include and pull from the flux extension?
        return all(k in item.get('obj').keys() for k in ["flux", "flux_err"])

    def is_valid_mask(self, item):
        if item.get('label') == 'None':
            return True
        if self.input_type == 'fits:hdulist':
            return self.hdu_is_valid_mask(item)
        if self.input_type == 'asdf:roman':
            return self.asdf_roman_is_valid_mask(item)
        if self.input_type == 'specutils:spectrum':
            return item.get('name') == 'mask'
        raise NotImplementedError("Unknown input type")

    def hdu_is_valid_mask(self, item):
        """
        Check if the HDU is valid to be imported for the mask in a Spectrum.

        Parameters
        ----------
        hdu : `astropy.io.fits.hdu.base.HDUBase`
            The HDU to check.

        Returns
        -------
        bool
            True if the HDU is a valid light curve HDU, False otherwise.
        """
        hdu = item.get('obj')
        return (len(getattr(hdu, 'shape', [])) == self.supported_flux_ndim
                and hdu.header.get('EXTNAME', '') == 'MASK')

    def asdf_roman_is_valid_mask(self, item):
        # TODO: should this instead have options just to include or not
        # include and pull from the flux extension?
        return False

    @observe('extension_selected',
             'unc_extension_selected',
             'mask_extension_selected')
    def _on_extension_change(self, change):
        self._clear_cache('spectrum')

    @cached_property
    def spectrum(self):
        if self.input_type == 'fits:hdulist':
            hdulist = self.input
            hdu = self.extension.selected_obj
            data = hdu.data
            header = hdu.header
            metadata = standardize_metadata(header)
            if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
                metadata[PRIHDR_KEY] = standardize_metadata(hdulist[0].header)

            wcs = WCS(header, hdulist)

            spectral_axis_index = None

            try:
                data_unit = u.Unit(header['BUNIT'])
            except Exception:
                data_unit = u.count

            if self.unc_extension.selected not in ('', 'None'):
                unc_hdu = self.unc_extension.selected_obj
                unc_data = unc_hdu.data
            else:
                unc_data = None
            if self.mask_extension.selected not in ('', 'None'):
                mask_hdu = self.mask_extension.selected_obj
                mask_data = mask_hdu.data
            else:
                mask_data = None

            if unc_data is not None:
                unc = StdDevUncertainty(unc_data * data_unit)
            else:
                unc = None

            if self.supported_flux_ndim == 2:
                spectral_axis_index = 1
                if data.shape[0] > data.shape[1]:
                    data = data.T
                    if unc_data is not None:
                        unc_data = unc_data.T
                    if mask_data is not None:
                        mask_data = mask_data.T
                    wcs = wcs.swapaxes(0, 1)
                    self.app.hub.broadcast(SnackbarMessage(
                        f"Transposed input data to {data.shape}",
                        sender=self, color="warning"))

            # Check for data types that have a GWCS stored in ASDF
            telescop = metadata[PRIHDR_KEY].get('TELESCOP', '').lower()
            exptype = metadata[PRIHDR_KEY].get('EXP_TYPE', '').lower()
            # NOTE: Alerted to deprecation of FILETYPE keyword from pipeline on 2022-07-08
            # Kept for posterity in for data processed prior to this date. Use EXP_TYPE instead
            filetype = metadata[PRIHDR_KEY].get('FILETYPE', '').lower()

            # Cubes and 2D data need different checks to see if we want WCS from ASDF
            asdf_3d = (telescop == 'jwst' and ('ifu' in exptype or
                                               'mrs' in exptype or
                                               filetype == '3d ifu cube'))
            asdf_2d = (wcs.world_axis_physical_types == [None, None] or
                       (hasattr(wcs, 'spectral') and wcs.spectral.naxis == 0))

            if (data.ndim == 2 and asdf_2d) or (data.ndim == 3 and asdf_3d):
                if 'ASDF' in hdulist:
                    try:
                        from stdatamodels import asdf_in_fits
                        tree = asdf_in_fits.open(hdulist).tree
                        if 'meta' in tree and 'wcs' in tree['meta']:
                            wcs = tree["meta"]["wcs"]
                            if isinstance(wcs, (list, tuple)):
                                wcs = wcs[0]
                            # Check needed for BSUB files, which we want to allow without worrying
                            # about the wavelength solution for now
                            if len(wcs.forward_transform.inputs) == 5:
                                wcs = None
                        else:
                            wcs = None
                    except ValueError:
                        wcs = None
                else:
                    wcs = None

            try:
                sc = Spectrum(flux=data * data_unit, uncertainty=unc,
                              mask=mask_data, meta=metadata, wcs=wcs,
                              spectral_axis_index=spectral_axis_index)
            except ValueError:
                # In some cases, the above call to Spectrum will fail if no
                # spectral axis is found in the WCS. Even without a spectral axis,
                # the Spectrum.read parser may work, so we try that next.
                # If that also fails, then drop the WCS.
                try:
                    sc = Spectrum.read(self._parser.input)
                except Exception:
                    # specutils.Spectrum reader would fail, so use no WCS
                    sc = Spectrum(
                            flux=data * data_unit, uncertainty=unc,
                            meta=metadata, spectral_axis_index=self.default_spectral_axis_index)
                else:
                    # raising an error here will consider this parser as non-valid
                    # so that specutils.Spectrum parser is preferred
                    raise

            # convert flux and uncertainty to per-pix2 if input is not a surface brightness
            target_flux_unit = None
            target_wave_unit = None
            apply_pix2 = 'FLUX' in self.extension.selected or 'ERR' in self.extension.selected
            flux = sc.flux
            if (apply_pix2 and
                    (not check_if_unit_is_per_solid_angle(flux.unit))):
                target_flux_unit = flux.unit / PIX2
            elif check_if_unit_is_per_solid_angle(flux.unit, return_unit=True) == "spaxel":
                # We need to convert spaxel to pixel squared, since spaxel isn't fully supported
                # by astropy
                # This is horribly ugly but just multiplying by u.Unit("spaxel") doesn't work
                target_flux_unit = flux.unit * u.Unit('spaxel') / PIX2
            if target_wave_unit is None and hdulist is not None:
                found_target = False
                for ext in ('SCI', 'FLUX', 'PRIMARY', 'DATA'):  # In priority order
                    if found_target:
                        break
                    if ext not in hdulist:
                        continue
                    hdr = hdulist[ext].header
                    # The WCS could be swapped or unswapped.
                    for cunit_num in (3, 1):
                        cunit_key = f"CUNIT{cunit_num}"
                        ctype_key = f"CTYPE{cunit_num}"
                        if cunit_key in hdr and 'WAV' in hdr[ctype_key]:
                            target_wave_unit = u.Unit(hdr[cunit_key])
                            found_target = True
                            break
            if target_wave_unit == sc.spectral_axis.unit:
                target_wave_unit = None
            if (target_wave_unit is None) and (target_flux_unit is None):  # Nothing to convert
                new_sc = sc
            elif target_flux_unit is None:  # Convert wavelength only
                new_sc = sc.with_spectral_axis_unit(target_wave_unit)
            elif target_wave_unit is None:  # Convert flux only and only PIX2 stuff
                new_sc = sc.with_flux_unit(target_flux_unit, equivalencies=_eqv_flux_to_sb_pixel())
            else:  # Convert both
                new_sc = sc.with_spectral_axis_and_flux_units(
                    target_wave_unit, target_flux_unit, flux_equivalencies=_eqv_flux_to_sb_pixel())
            if target_wave_unit is not None:
                new_sc.meta['_orig_spec'] = sc
            # Since we create a new Spectrum, we need to copy over any original WCS info
            # since the WCS will be replaced by a SpectralGWCS object instead of the original
            # astropy.wcs.WCS object.
            # This is needed for the subset tools to work properly.
            if new_sc.flux.ndim == 3 and _get_celestial_wcs(sc.wcs) is not None:
                new_sc.meta['_orig_spatial_wcs'] = _get_celestial_wcs(sc.wcs)

            return new_sc
        elif self.input_type == 'asdf:roman':
            def _to_unit(x):
                """Coerce str/bytes/Unit to astropy.units.Unit."""
                if isinstance(x, bytes):
                    x = x.decode()
                return u.Unit(x)

            roman = self.input["roman"]
            meta = roman["meta"]
            data = roman["data"]
            extension = self.extension.selected_obj
            if self.supported_flux_ndim == 1:
                wavelength = np.asarray(extension["wl"])
                flux = np.asarray(extension["flux"])
                wl_unit = _to_unit(meta["unit_wl"])
                flux_unit = _to_unit(meta["unit_flux"])

                # TODO: expose option in unc_extension that defaults to pulling
                # from flux extension, but also allowing user to set to "None"
                # to skip loading uncertainty
                flux_error = extension.get("flux_error", None)
                variance = extension.get("var", None)
                uncertainty = None
            elif self.supported_flux_ndim == 2:
                # TODO: handle detecting/selecting spectral axis?
                flux = np.asarray(extension["spectrum"]).transpose()
                flux_unit = _to_unit(meta["unit_flux"])
                if extension['wavelength'] is not None:
                    wavelength = np.asarray(extension["wavelength"])
                    wl_unit = _to_unit(meta["unit_wl"])
                else:
                    wavelength = np.arange(flux.shape[1])
                    wl_unit = u.pix

                flux_error = None
                variance = extension.get("variance", None)
                if variance is not None:
                    variance = np.asarray(variance).transpose()
                uncertainty = None
            else:
                raise NotImplementedError("Only 1D and 2D spectra are supported for ASDF Roman data")  # noqa

            if flux_error is not None:
                uncertainty = StdDevUncertainty(np.asarray(flux_error) * flux_unit)
            elif variance is not None:
                var = np.asarray(variance) * (flux_unit ** 2)
                var = np.where(np.asarray(var.value) < 0, np.nan, var.value) * var.unit
                uncertainty = StdDevUncertainty(np.sqrt(var))
            else:
                uncertainty = None

            spectrum = Spectrum(
                flux=flux * flux_unit,
                spectral_axis=wavelength * wl_unit,
                uncertainty=uncertainty
            )
            return spectrum
        elif self.input_type == 'specutils:spectrum':
            spectrum = self.input
            # TODO: remove uncertainty or mask if requested
            return spectrum



import traceback as tb

import astropy.units as u
from glue.core.message import Message

__all__ = ['NewViewerMessage', 'ViewerAddedMessage', 'ViewerRemovedMessage', 'LoadDataMessage',
           'AddDataMessage', 'SnackbarMessage', 'RemoveDataMessage', 'SubsetRenameMessage',
           'ViewerVisibleLayersChangedMessage', 'LayersFinalizedMessage',
           'AddLineListMessage', 'RowLockMessage',
           'SliceSelectSliceMessage', 'SliceValueUpdatedMessage',
           'SliceToolStateMessage',
           'CatalogResultsChangedMessage', 'CatalogSelectClickEventMessage',
           'FootprintSelectClickEventMessage', 'FootprintOverlayClickMessage',
           'TableClickMessage', 'LinkUpdatedMessage', 'ExitBatchLoadMessage',
           'AstrowidgetMarkersChangedMessage', 'MarkersPluginUpdate',
           'GlobalDisplayUnitChanged', 'ChangeRefDataMessage',
           'PluginTableAddedMessage', 'PluginTableModifiedMessage',
           'PluginPlotAddedMessage', 'PluginPlotModifiedMessage',
           'IconsUpdatedMessage', 'RestoreToolbarMessage']


class NewViewerMessage(Message):
    """Message to trigger viewer creation in the application."""
    def __init__(self, cls, data, x_attr=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._cls = cls
        self._data = data
        self._x_attr = x_attr

    @property
    def cls(self):
        return self._cls

    @property
    def data(self):
        return self._data

    @property
    def x_attr(self):
        return self._x_attr


class ViewerAddedMessage(Message):
    """Unlike `NewViewerMessage`, this should be emitted after a viewer is created."""
    def __init__(self, viewer_id, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # NOT viewer reference, which is not unique and can be None
        self._viewer_id = viewer_id

    @property
    def viewer_id(self):
        return self._viewer_id


class ViewerRemovedMessage(Message):
    """Message emitted after a viewer is destroyed by the application."""
    def __init__(self, viewer_id, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # NOT viewer reference, which is not unique and can be None
        self._viewer_id = viewer_id

    @property
    def viewer_id(self):
        return self._viewer_id


class ViewerRenamedMessage(Message):
    """Message emitted after a viewer is destroyed by the application."""
    def __init__(self, old_viewer_ref, new_viewer_ref, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._old_viewer_ref = old_viewer_ref
        self._new_viewer_ref = new_viewer_ref

    @property
    def old_viewer_ref(self):
        return self._old_viewer_ref

    @property
    def new_viewer_ref(self):
        return self._new_viewer_ref


class LoadDataMessage(Message):
    def __init__(self, path, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._path = path

    @property
    def path(self):
        return self._path


class LayersFinalizedMessage(Message):
    """
    This event type is a temporary solution to AddDataMessage ordering issues.
    In app.py after the AddDataMessage is broadcast, modifications are made to
    the viewer layers (e.g. setting visibility). This message is broadcast after
    those modifications are made so that plugins can also set layer visiblity
    after data is added to a viewer without the visibility being modified by the
    code in app.py that runs after the AddDataMessage is broadcast."""

    def __init__(self, data, viewer, viewer_id=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._data = data
        self._viewer = viewer
        self._viewer_id = viewer_id

    @property
    def data(self):
        return self._data

    @property
    def viewer(self):
        return self._viewer

    @property
    def viewer_id(self):
        return self._viewer_id


class AddDataMessage(Message):
    """
    Emitted AFTER data is added to a viewer
    """
    def __init__(self, data, viewer, viewer_id=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._data = data
        self._viewer = viewer
        self._viewer_id = viewer_id

    @property
    def data(self):
        return self._data

    @property
    def viewer(self):
        return self._viewer

    @property
    def viewer_id(self):
        return self._viewer_id


class RemoveDataMessage(Message):
    def __init__(self, data, viewer, viewer_id=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._data = data
        self._viewer = viewer
        self._viewer_id = viewer_id

    @property
    def data(self):
        return self._data

    @property
    def viewer(self):
        return self._viewer

    @property
    def viewer_id(self):
        return self._viewer_id


class ChangeRefDataMessage(Message):
    def __init__(self, data, viewer, viewer_id=None, old=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._data = data
        self._viewer = viewer
        self._viewer_id = viewer_id
        self._old = old

    @property
    def data(self):
        return self._data

    @property
    def viewer(self):
        return self._viewer

    @property
    def viewer_id(self):
        return self._viewer_id

    @property
    def old(self):
        return self._old


class SubsetRenameMessage(Message):
    def __init__(self, subset_group, old_label, new_label, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._subset_group = subset_group
        self._old_label = old_label
        self._new_label = new_label

    @property
    def subset_group(self):
        return self._subset_group

    @property
    def old_label(self):
        return self._old_label

    @property
    def new_label(self):
        return self._new_label


class SnackbarMessage(Message):
    def __init__(self, text, color=None, timeout=5000, loading=False,
                 traceback=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._text = text
        self._color = color
        self._timeout = timeout
        self._loading = loading
        self._traceback = tb.format_exc() if isinstance(traceback, Exception) else traceback  # noqa

    @property
    def text(self):
        return self._text

    @property
    def color(self):
        return self._color

    @property
    def timeout(self):
        return self._timeout

    @property
    def loading(self):
        return self._loading

    @property
    def traceback(self):
        return self._traceback


class ConfigurationLoadedMessage(Message):
    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._config = config

    @property
    def config(self):
        return self._config


class AddDataToViewerMessage(Message):
    """
    Emitted to request data is added to a viewer (BEFORE the data is actually added)
    """
    def __init__(self, viewer_reference, data_label, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._viewer_reference = viewer_reference
        self._data_label = data_label

    @property
    def viewer_reference(self):
        return self._viewer_reference

    @property
    def data_label(self):
        return self._data_label


class RemoveDataFromViewerMessage(Message):
    def __init__(self, viewer_reference, data_label, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._viewer_reference = viewer_reference
        self._data_label = data_label

    @property
    def viewer_reference(self):
        return self._viewer_reference

    @property
    def data_label(self):
        return self._data_label


class ViewerVisibleLayersChangedMessage(Message):
    def __init__(self, viewer_reference, visible_layers, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._viewer_reference = viewer_reference
        self._visible_layers = visible_layers

    @property
    def viewer_reference(self):
        return self._viewer_reference

    @property
    def visible_layers(self):
        return self._visible_layers


class AddLineListMessage(Message):
    def __init__(self, table, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._table = table

    @property
    def table(self):
        return self._table


class LineIdentifyMessage(Message):
    def __init__(self, name_rest, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name_rest = name_rest

    @property
    def name_rest(self):
        return self._name_rest


class SpectralMarksChangedMessage(Message):
    def __init__(self, marks, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._marks = marks

    @property
    def names_rest(self):
        return [m.name_rest for m in self.marks]

    @property
    def marks(self):
        return self._marks


class CatalogResultsChangedMessage(Message):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class FootprintMarkVisibilityChangedMessage(Message):
    """Message to notify that footprint overlays have changed."""

    def __init__(self, viewer_id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._viewer_id = viewer_id

    @property
    def viewer_id(self):
        return self._viewer_id


class CatalogSelectClickEventMessage(Message):
    def __init__(self, x, y, *args, **kwargs):
        self.x = x
        self.y = y
        super().__init__(*args, **kwargs)


class FootprintSelectClickEventMessage(Message):
    """
    Message emitted when a user clicks on a viewer to select an overlay.
    """

    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x = data["domain"]["x"]
        self.y = data["domain"]["y"]


class FootprintOverlayClickMessage(Message):
    """
    Message emitted when a user clicks on a viewer to select a footprint/region overlay.
    """

    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x = data["domain"]["x"]
        self.y = data["domain"]["y"]


class RedshiftMessage(Message):
    '''Messages related to Specviz redshift slider'''
    def __init__(self, param, value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._param = param
        self._value = value

    @property
    def param(self):
        return self._param

    @property
    def value(self):
        return self._value


class RowLockMessage(Message):
    def __init__(self, is_locked, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_locked = is_locked

    @property
    def is_locked(self):
        return self._is_locked


class TableClickMessage(Message):
    '''Message generated by Mosviz table to zoom to object on image'''
    def __init__(self, selected_index, shared_image=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._selected_index = selected_index
        self._shared_image = shared_image

    @property
    def selected_index(self):
        return self._selected_index

    @property
    def shared_image(self):
        return self._shared_image


class SliceSelectSliceMessage(Message):
    '''Message generated by the cubeviz helper and processed by the slice plugin to sync
    slice selection across all viewers'''
    def __init__(self, value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = value

    @property
    def value(self):
        return self._value


class SliceValueUpdatedMessage(Message):
    '''Message generated by the slice plugin when the selected slice is updated'''
    def __init__(self, value, value_unit, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = value
        self.value_unit = value_unit


class SliceToolStateMessage(Message):
    '''Message generated by the select slice plot plugin when activated/deactivated'''
    def __init__(self, change, viewer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._change = change
        self._viewer = viewer

    @property
    def change(self):
        return self._change

    @property
    def viewer(self):
        return self._viewer


class LinkUpdatedMessage(Message):
    '''Message generated when the WCS/pixel linking is changed'''
    def __init__(self, link_type, wcs_use_fallback, wcs_fast_approximation, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._link_type = link_type
        self._wcs_use_fallback = wcs_use_fallback
        self._wcs_fast_approximation = wcs_fast_approximation

    @property
    def link_type(self):
        return self._link_type

    @property
    def wcs_use_fallback(self):
        return self._wcs_use_fallback

    @property
    def wcs_fast_approximation(self):
        return self._wcs_fast_approximation


class ExitBatchLoadMessage(Message):
    '''Message generated when exiting the outermost batch_load context manager'''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AstrowidgetMarkersChangedMessage(Message):
    '''Message generated when markers are added/removed from an image viewer via astrowidgets API'''
    def __init__(self, has_markers, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_markers = has_markers

    @property
    def has_markers(self):
        return self._has_markers


class MarkersPluginUpdate(Message):
    '''Message when the length of the markers plugin table changes'''
    def __init__(self, table_length, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.table_length = table_length


class GlobalDisplayUnitChanged(Message):
    '''Message generated when the (x or y axis) unit of the spectrum viewer is
    changed, which is used app-wide to inform display units that depend on the
    unit choice and flux<>sb toggle of the spectrum viewer.'''

    def __init__(self, axis, unit, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._axis = axis
        self._unit = unit

    @property
    def axis(self):
        return self._axis

    @property
    def unit(self):
        return u.Unit(self._unit)


class PluginTableAddedMessage(Message):
    '''Message generated when a plugin table is initialized'''
    def __init__(self, sender):
        super().__init__(sender)

    @property
    def table(self):
        return self.sender

    @property
    def plugin(self):
        return self.sender._plugin


class PluginTableModifiedMessage(PluginTableAddedMessage):
    '''Message generated when the items in a plugin table are changed'''
    def __init__(self, sender):
        super().__init__(sender)


class PluginPlotAddedMessage(Message):
    '''Message generated when a plugin plot is initialized'''
    def __init__(self, sender):
        super().__init__(sender)

    @property
    def plot(self):
        return self.sender

    @property
    def plugin(self):
        return self.sender._plugin


class PluginPlotModifiedMessage(PluginPlotAddedMessage):
    '''Message generated when the items in a plugin plot are changed'''
    def __init__(self, sender):
        super().__init__(sender)


class IconsUpdatedMessage(Message):
    '''Message generated when the viewer or layer icons are updated'''
    def __init__(self, icon_type, icons, **kwargs):
        # icon_type = 'layer' or 'viewer'
        super().__init__(**kwargs)
        self.icon_type = icon_type
        # icons might be a CallbackDict, cast to ensure its a dictionary
        self.icons = dict(icons)


class RestoreToolbarMessage(Message):
    '''Message generated to restore all toolbar instances to their original configuration'''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



"""
Tests for the parsers module.
"""
from collections import defaultdict
import re

import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import StdDevUncertainty
from glue.core import Data
from specutils import Spectrum, SpectrumList

from jdaviz.configs.specviz.plugins.parsers import (
    specviz_spectrum1d_parser,
    group_spectra_by_filename,
    combine_lists_to_1d_spectrum,
    split_spectrum_with_2D_flux_array
)


class TestSpecvizSpectrum1DParser:
    """
    Test the main specviz_spectrum1d_parser function.
    """
    @pytest.mark.parametrize('label', [None, 'custom_label'])
    def test_parse_single_spectrum1d(self, specviz_helper, spectrum1d, label):
        """
        Test parsing a single Spectrum1D object.
        """
        specviz_spectrum1d_parser(specviz_helper.app, spectrum1d, data_label=label)

        assert len(specviz_helper.app.data_collection) == 1
        if label:
            assert specviz_helper.app.data_collection[0].label == label
        else:
            assert specviz_helper.app.data_collection[0].label == 'Spectrum'

    def test_parse_spectrum_list(self, specviz_helper, premade_spectrum_list):
        """
        Test parsing a SpectrumList.
        """
        specviz_spectrum1d_parser(specviz_helper.app,
                                  premade_spectrum_list,
                                  data_label=[f"spec{i}_label"
                                              for i in range(len(premade_spectrum_list))])

        assert len(specviz_helper.app.data_collection) == len(premade_spectrum_list)
        for i, data in enumerate(specviz_helper.app.data_collection):
            assert data.label == f"spec{i}_label"

    def test_parse_2d_flux_spectrum(self, specviz_helper, spectrum2d):
        """
        Test that 2D flux spectra are split automatically.
        """
        specviz_spectrum1d_parser(specviz_helper.app, spectrum2d, data_label='spec2d_label')

        assert len(specviz_helper.app.data_collection) == np.shape(spectrum2d)[0]
        for i, data in enumerate(specviz_helper.app.data_collection):
            assert data.label == f"spec2d_label [{i}]"

    def test_parse_spectrum_collection(self, specviz_helper, spectrum_collection):
        """
        Test parsing a SpectrumCollection.
        """
        msg = 'SpectrumCollection detected. Please provide a Spectrum or SpectrumList'
        with pytest.raises(TypeError, match=msg):
            specviz_spectrum1d_parser(specviz_helper.app, spectrum_collection)

    def test_parse_mismatched_label_length(self, specviz_helper, premade_spectrum_list):
        """
        Test error when label list length doesn't match data length.
        """
        msg = "Length of data labels list (1) is different than length of list of data (5)"
        with pytest.raises(ValueError, match=re.escape(msg)):
            specviz_spectrum1d_parser(specviz_helper.app,
                                      premade_spectrum_list,
                                      data_label=['oops'])

    def test_parse_concat_by_file(self, specviz_helper):
        """
        Test the concat_by_file functionality.
        """
        spec1 = Spectrum(
            spectral_axis=np.linspace(5000, 5500, 50) * u.AA,
            flux=np.random.random(50) * u.Jy,
            meta={'header': {'FILENAME': 'same_file.fits'}}
        )
        spec2 = Spectrum(
            spectral_axis=np.linspace(5500, 6000, 50) * u.AA,
            flux=np.random.random(50) * u.Jy,
            meta={'header': {'FILENAME': 'same_file.fits'}}
        )
        spec_list = SpectrumList([spec1, spec2])

        specviz_spectrum1d_parser(specviz_helper.app,
                                  spec_list,
                                  data_label='test_concat',
                                  concat_by_file=True)

        # Should have 2 individual spectra + 1 combined
        assert len(specviz_helper.app.data_collection) == 3
        labels = [
            d.label for d in specviz_helper.app.data_collection
        ]
        assert any('Combined' in label for label in labels)

    def test_parse_file_path(self, specviz_helper, tmp_path):
        """
        Test parsing from a file path.
        """
        spec = Spectrum(
            spectral_axis=np.linspace(5000, 6000, 100) * u.AA,
            flux=np.random.random(100) * u.Jy,
            uncertainty=StdDevUncertainty(
                np.random.random(100) * 0.1 * u.Jy
            )
        )

        filepath = tmp_path / 'test_spectrum.fits'
        spec.write(str(filepath), overwrite=True)

        specviz_spectrum1d_parser(specviz_helper.app,
                                  str(filepath),
                                  data_label='test_filepath')

        assert len(specviz_helper.app.data_collection) == 1

    def test_parse_metadata_standardization(self, specviz_helper):
        """
        Test that metadata is standardized.
        """
        spec = Spectrum(
            spectral_axis=np.linspace(5000, 6000, 50) * u.AA,
            flux=np.random.random(50) * u.Jy,
            meta={'TELESCOPE': 'TEST', 'INSTRUMENT': 'FAKE'}
        )

        specviz_spectrum1d_parser(specviz_helper.app, spec, data_label='test_meta')

        loaded_spec = specviz_helper.get_data('test_meta')
        assert 'TELESCOPE' in loaded_spec.meta

    def test_parse_show_in_viewer_false(self, specviz_helper, spectrum1d):
        """
        Test that show_in_viewer=False doesn't add to viewer.
        """
        specviz_spectrum1d_parser(specviz_helper.app,
                                  spectrum1d,
                                  data_label='hidden_spec',
                                  show_in_viewer=False)

        assert len(specviz_helper.app.data_collection) == 1

        viewer = specviz_helper.app.get_viewer(
            specviz_helper._default_spectrum_viewer_reference_name
        )
        # Check that no data is shown in viewer
        assert len(viewer.data()) == 0


class TestGroupSpectraByFilename:
    """
    Test the group_spectra_by_filename function.
    """
    def test_group_single_filename(self):
        """
        Test grouping when all spectra have the same filename.
        """
        data1 = Data(label='spec1')
        data1.meta = {'FILENAME': 'test_file.fits'}
        data2 = Data(label='spec2')
        data2.meta = {'FILENAME': 'test_file.fits'}
        data3 = Data(label='spec3')
        data3.meta = {'FILENAME': 'test_file.fits'}

        datasets = [data1, data2, data3]

        result = group_spectra_by_filename(datasets)

        assert 'test_file.fits' in result
        assert len(result['test_file.fits']) == 3

    def test_group_multiple_filenames(self):
        """
        Test grouping when spectra come from different files.
        """
        data1 = Data(label='spec1')
        data1.meta = {'FILENAME': 'file1.fits'}
        data2 = Data(label='spec2')
        data2.meta = {'FILENAME': 'file2.fits'}
        data3 = Data(label='spec3')
        data3.meta = {'FILENAME': 'file1.fits'}

        datasets = [data1, data2, data3]

        result = group_spectra_by_filename(datasets)

        assert len(result) == 2
        assert len(result['file1.fits']) == 2
        assert len(result['file2.fits']) == 1

    def test_group_missing_filename(self):
        """
        Test grouping when some spectra don't have FILENAME metadata.
        """
        data1 = Data(label='spec1')
        data1.meta = {'FILENAME': 'test.fits'}
        data2 = Data(label='spec2')
        data2.meta = {}
        data3 = Data(label='spec3')
        data3.meta = {'FILENAME': 'test.fits'}

        datasets = [data1, data2, data3]

        result = group_spectra_by_filename(datasets)

        assert 'test.fits' in result
        assert None in result
        assert len(result['test.fits']) == 2
        assert len(result[None]) == 1

    def test_group_empty_datasets(self):
        """
        Test grouping with empty dataset list.
        """
        datasets = []

        result = group_spectra_by_filename(datasets)

        assert isinstance(result, defaultdict)
        assert len(result) == 0


# Copied and pasted from test_spectrum_list.py
@pytest.mark.parametrize('with_uncertainty', [True, False])
def test_combine_lists_to_1d_spectrum(with_uncertainty):
    wl = [1, 2, 3] * u.nm
    fnu = [10, 20, 30] * u.Jy
    if with_uncertainty:
        dfnu = [4, 5, 6] * u.Jy
    else:
        dfnu = None

    spec = combine_lists_to_1d_spectrum(wl, fnu, dfnu, u.nm, u.Jy)
    assert isinstance(spec, Spectrum)
    assert isinstance(spec.flux, u.Quantity)
    assert spec.flux.unit == u.Jy

    assert isinstance(spec.spectral_axis, u.Quantity)
    assert spec.spectral_axis.unit == u.nm

    assert np.all(spec.flux.value == np.array([10, 20, 30]))
    assert np.all(spec.spectral_axis.value == np.array([1, 2, 3]))

    if with_uncertainty:
        assert isinstance(spec.uncertainty, StdDevUncertainty)
        assert np.all(spec.uncertainty.array == np.array([4, 5, 6]))
    else:
        assert spec.uncertainty is None


class TestSplitSpectrumWith2DFluxArray:
    """
    Test the split_spectrum_with_2D_flux_array function.
    """
    def test_split_2d_flux_basic(self, spectrum2d):
        """
        Test splitting a spectrum with 2D flux array into a list.
        """
        result = split_spectrum_with_2D_flux_array(spectrum2d)

        assert len(result) == 5
        for i, sub_spec in enumerate(result):
            assert sub_spec.flux.shape == (spectrum2d.flux.shape[1],)
            assert np.array_equal(sub_spec.flux.value, spectrum2d.flux[i, :].value)
            assert np.array_equal(
                sub_spec.spectral_axis.value,
                spectrum2d.spectral_axis.value
            )

        # No uncertainty or mask in input, so none in output
        for sub_spec in result:
            assert sub_spec.uncertainty is None
            assert sub_spec.mask is None

    def test_split_2d_flux_with_uncertainty(self, spectrum2d):
        """
        Test splitting a spectrum with uncertainty.
        """
        uncertainty_2d = StdDevUncertainty(
            np.random.random(spectrum2d.shape) * spectrum2d.flux.unit,
        )
        spec = Spectrum(
            spectral_axis=spectrum2d.spectral_axis,
            flux=spectrum2d.flux,
            uncertainty=uncertainty_2d
        )

        result = split_spectrum_with_2D_flux_array(spec)

        assert len(result) == spectrum2d.flux.shape[0]
        for i, sub_spec in enumerate(result):
            assert sub_spec.uncertainty is not None
            assert np.array_equal(
                sub_spec.uncertainty.array,
                uncertainty_2d.array[i, :]
            )

    def test_split_2d_flux_with_mask(self, spectrum2d):
        """
        Test splitting a spectrum with mask.
        """
        mask_2d = np.zeros(spectrum2d.flux.shape, dtype=bool)
        mask_2d[:, -5:] = True
        spec = Spectrum(
            spectral_axis=spectrum2d.spectral_axis,
            flux=spectrum2d.flux,
            mask=mask_2d
        )

        result = split_spectrum_with_2D_flux_array(spec)

        assert len(result) == spectrum2d.flux.shape[0]
        for i, sub_spec in enumerate(result):
            assert sub_spec.mask is not None
            assert np.array_equal(sub_spec.mask, mask_2d[i, :])

    def test_split_2d_flux_with_metadata(self, spectrum2d):
        """
        Test that metadata is preserved when splitting.
        """
        meta = {'TELESCOPE': 'TEST', 'INSTRUMENT': 'FAKE'}
        spec = Spectrum(
            spectral_axis=spectrum2d.spectral_axis,
            flux=spectrum2d.flux,
            meta=meta
        )

        result = split_spectrum_with_2D_flux_array(spec)

        for sub_spec in result:
            assert sub_spec.meta == meta



import pytest
from unittest.mock import PropertyMock, patch
import warnings

import numpy as np
from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.nddata import CCDData, NDDataArray
from glue.core import ComponentID
from glue.core.roi import CircularROI
from specutils import SpectralRegion, Spectrum

from jdaviz.core.helpers import _next_subset_num
from jdaviz.core.loaders import ObjectResolver, ObjectParser


class MockGroupItem:
    def __init__(self, label):
        self.label = label


@pytest.mark.parametrize(
    ('label', 'prefix', 'answer'),
    [('Subset 1', 'Subset', 2),
     ('Subset 10', 'Subset', 11),
     ('MaskedSubset', 'MaskedSubset', 1),
     ('acs_47tuc_1[SCI,1]', 'acs_47tuc_1[SCI,1]', 1),
     ('acs_47tuc_1[SCI,1] 1', 'acs_47tuc_1[SCI,1]', 2),
     ('asdakdh663 dada233 32432 2', 'asdakdh663', 3),
     ('asdakdh663 dada233 32432 2', 'hmmm', 1),
     ('42 1337 kookookachu', '42', 1)])
def test_next_subset_num(label, prefix, answer):
    mocked_group = [MockGroupItem(label)]
    assert _next_subset_num(prefix, mocked_group) == answer


class TestConfigHelperSpec:
    @pytest.fixture(autouse=True)
    def _setup_class(self, deconfigged_helper, spectrum1d, spectrum2d, multi_order_spectrum_list):
        self.config_helper = deconfigged_helper

        self.spec = spectrum1d
        self.label = "Test 1D Spectrum"
        spectral_axis_unit = u.AA

        self.spec2 = spectrum1d._copy(
            spectral_axis=spectrum1d.spectral_axis + 1000 * spectral_axis_unit)
        self.label2 = "Test 1D Spectrum 2"

        self.config_helper.load(spectrum1d, data_label=self.label, format='1D Spectrum')
        self.config_helper.load(self.spec2, data_label=self.label2, format='1D Spectrum')

        # Add 3 subsets to cover different parts of spec and spec2
        self.config_helper.plugins['Subset Tools'].import_region(
            SpectralRegion(6000*spectral_axis_unit, 6500*spectral_axis_unit))
        # combination_mode = 'new' is the default, so no need to specify
        self.config_helper.plugins['Subset Tools'].import_region(
            SpectralRegion(6700*spectral_axis_unit, 7200*spectral_axis_unit))
        self.config_helper.plugins['Subset Tools'].import_region(
            SpectralRegion(8200*spectral_axis_unit, 8800*spectral_axis_unit))

        self.data = self.config_helper.app.data_collection[self.label]

    @pytest.mark.parametrize(
        ('label', 'subset_name', 'answer'),
        [('Test 1D Spectrum', 'Subset 1',
          [False, False, False, True, True, True, True, True, True, True]),
         ('Test 1D Spectrum', 'Subset 2',
          [True, True, True, True, False, False, True, True, True, True]),
         ('Test 1D Spectrum', 'Subset 3',
          [True, True, True, True, True, True, True, True, True, True]),
         ('Test 1D Spectrum 2', 'Subset 1',
          [True, True, True, True, True, True, True, True, True, True]),
         ('Test 1D Spectrum 2', 'Subset 2',
          [False, True, True, True, True, True, True, True, True, True]),
         ('Test 1D Spectrum 2', 'Subset 3',
          [True, True, True, True, True, True, False, False, False, True])])
    def test_get_data_with_one_subset_per_data(self, label, subset_name, answer):
        results = self.config_helper._get_data(data_label=label, spectral_subset=subset_name)
        assert list(results.mask) == answer

    def test_get_data_errors(self):
        # multiple data, no label
        with pytest.raises(ValueError, match='data_label must be set if more'):
            self.config_helper._get_data()

        # data label not in data collection
        with pytest.raises(ValueError, match='Blah not in '):
            self.config_helper._get_data(data_label="Blah")

        # invalid cls type
        self.config_helper.app.data_collection.remove(
            self.config_helper.app.data_collection[self.label2])
        with pytest.raises(TypeError, match="cls in get_data must be a class or None."):
            self.config_helper._get_data('Test 1D Spectrum', cls=42)

        # spectral subset name not in data collection
        with pytest.raises(ValueError, match="not in list of valid subset names"):
            self.config_helper._get_data('Test 1D Spectrum', spectral_subset="Fail")

    def test_get_data_no_label_one_in_dc(self):
        self.config_helper.app.data_collection.remove(
            self.config_helper.app.data_collection[self.label2])
        results = self.config_helper._get_data()
        assert_quantity_allclose(results.flux,
                                 self.spec.flux, atol=1e-5 * u.Unit(self.spec.flux.unit))

    def test_get_clone_viewer_reference(self):
        """
        Test _get_clone_viewer_reference alone and with multiple existing clones.
        """
        base_viewer_name = '1D Spectrum'
        result = self.config_helper._get_clone_viewer_reference(base_viewer_name)
        assert result == f"{base_viewer_name}[1]"

        # Get the original viewers dict
        original_viewers = self.config_helper.viewers
        mock_viewers = dict(original_viewers)
        for i in range(1, 4):
            mock_viewers[f"{base_viewer_name}[{i}]"] = original_viewers[base_viewer_name]

        # Patch the viewers property to return our mocked dict
        with patch.object(type(self.config_helper),
                          'viewers',
                          new_callable=PropertyMock,
                          return_value=mock_viewers):
            result = self.config_helper._get_clone_viewer_reference(base_viewer_name)
            assert result == f"{base_viewer_name}[4]"

    def test_set_data_component(self):
        """
        Test _set_data_component with a new component label and update it.
        """
        original_n_components = len(self.data.components)

        new_values = np.ones(len(self.data.get_object().flux))
        self.config_helper._set_data_component(self.data, 'test_component', new_values)

        assert len(self.data.components) == original_n_components + 1
        assert 'test_component' in self.config_helper._component_ids

        # Now check updating existing component
        n_components_after_add = len(self.data.components)
        # Now update it
        updated_values = np.ones(len(self.data.get_object().flux)) * 2
        self.config_helper._set_data_component(self.data, 'test_component', updated_values)

        # Should not add a new component, just update
        assert len(self.data.components) == n_components_after_add

    def test_set_data_component_existing_in_data_not_in_cache(self):
        """
        Test _set_data_component with component in data but not cache.
        """
        # Add a component directly to data, bypassing the cache
        comp_id = ComponentID('direct_component')
        values = np.ones(len(self.data.get_object().flux))
        self.data.add_component(values, comp_id)

        # Now use _set_data_component with the same label
        updated_values = np.ones(len(self.data.get_object().flux)) * 3
        self.config_helper._set_data_component(self.data, 'direct_component', updated_values)

        # Should update, not add a duplicate
        component_labels = [c.label for c in self.data.components]
        assert component_labels.count('direct_component') == 1

    def test_set_data_component_cached_component_id(self):
        """
        Test _set_data_component uses cached ComponentID.
        """
        # Add a component to populate the cache
        values = np.ones(len(self.data.get_object().flux))
        self.config_helper._set_data_component(self.data, 'cached_comp', values)

        # Get the cached component ID
        cached_id = self.config_helper._component_ids['cached_comp']

        # Update the component
        new_values = np.ones(len(self.data.get_object().flux)) * 5
        self.config_helper._set_data_component(self.data, 'cached_comp', new_values)

        # Should still use the same component ID
        assert self.config_helper._component_ids['cached_comp'] == cached_id


class TestConfigHelperSubsets:
    @pytest.fixture(autouse=True)
    def setup_class(self, cubeviz_helper, deconfigged_helper, image_cube_hdu_obj):
        """
        Set up test data with spatial subsets.
        """
        # self.config_helper = deconfigged_helper
        self.config_helper = cubeviz_helper
        # self.config_helper.load(image_cube_hdu_obj, format='Spectral Cube')
        self.config_helper.load_data(image_cube_hdu_obj)
        self.data = self.config_helper.app.data_collection[0]
        self.label = self.data.label

        subset_plugin = self.config_helper.plugins['Subset Tools']
        # Subset 1, Spatial
        subset_plugin.import_region(CircularROI(5, 5, 3))
        # Subset 2, Spectral
        subset_plugin.import_region(SpectralRegion(4.6e-7 * u.m, 4.8e-7 * u.m))

    @pytest.mark.parametrize(
        ('spatial_subset', 'spectral_subset', 'temporal_subset', 'mask_subset'),
        [('Subset 1', None, None, None),
         (None, 'Subset 2', None, None),
         (None, None, 'Subset 1', None),
         (None, None, None, 'Subset 2'),
         ('Subset 1', 'Subset 2', None, None)])
    def test_get_data_spatial_spectral_subsets(self,
                                               spatial_subset, spectral_subset,
                                               temporal_subset, mask_subset):
        """
        Test _get_data spatial subset code path
        (exception handling is difficult to test without deep mocking).
        """
        # Verify the spatial subset code path executes without error
        result = self.config_helper._get_data(data_label=self.label,
                                              spatial_subset=spatial_subset,
                                              spectral_subset=spectral_subset,
                                              temporal_subset=temporal_subset,
                                              mask_subset=mask_subset,
                                              cls=Spectrum)

        # Verify we got a result
        assert isinstance(result, Spectrum)

        spatial_spectral_combo = spatial_subset is not None and spectral_subset is not None
        if mask_subset is not None or temporal_subset is not None or spatial_spectral_combo:
            # Verify that result is a Spectrum with a mask applied
            assert hasattr(result, 'mask')
            assert result.mask is not None
            # The mask should have some True values (masked regions)
            assert np.any(result.mask)

    # TODO: May have to change cls=Spectrum to something else
    #  once we adjust the logic for deconfigged
    @pytest.mark.parametrize(
        ('spatial_subset', 'spectral_subset', 'temporal_subset', 'mask_subset', 'cls',
         'error_type', 'error_msg'), [
            (None, 'Subset 2', None, 'Subset 2', Spectrum,
             ValueError, 'cannot use both mask_subset and spectral_subset'),
            (None, None, 'Subset 1', 'Subset 2', Spectrum,
             ValueError, 'cannot use both mask_subset and temporal_subset'),
            ('Subset 1', None, None, None, None,
             AttributeError, 'A valid cls must be provided to apply subset'),
            (None, None, None, 'Subset 2', None,
             AttributeError, 'A valid cls must be provided to apply subset'),
            ('Subset 2', None, None, None, Spectrum,
             ValueError, 'is not a spatial subset'),
            (None, 'Subset 1', None, None, Spectrum,
             ValueError, 'is not a spectral subset')])
    def test_get_data_subset_errors(self,
                                    spatial_subset, spectral_subset,
                                    temporal_subset, mask_subset, cls,
                                    error_type, error_msg):
        if cls is None:
            # Injecting a 'Trace' key in meta is the only way to
            # get past the cls check and into the subset application code
            self.data.meta['Trace'] = True
            if '_native_data_cls' in self.data.meta:
                self.data.meta.pop('_native_data_cls')

        with pytest.raises(error_type, match=error_msg):
            self.config_helper._get_data(data_label=self.label,
                                         spatial_subset=spatial_subset,
                                         spectral_subset=spectral_subset,
                                         temporal_subset=temporal_subset,
                                         mask_subset=mask_subset,
                                         cls=cls)


@pytest.mark.skip(reason="TODO: need to adjust logic for deconfigged (lines 611-626)")
@pytest.mark.parametrize('data_tuple',
                         [('mos_image', 'Image', CCDData),
                          ('mos_spectrum1d', '1D Spectrum', Spectrum),
                          ('mos_spectrum2d', '2D Spectrum', Spectrum),
                          ('spectral_cube_wcs', 'Spectral Cube', Spectrum),
                          ('image_cube_hdu_obj', 'Spectral Cube', Spectrum),
                          # TODO: enable when rampviz loader is ready
                          # ('jwst_level_1b_ramp', 'Ramp', NDDataArray),
                          ])
def test_get_data_cls(deconfigged_helper, request, data_tuple):
    """
    Test _get_data cls inference.
    1) CCDData -> CCDData
    2) 1D Spectrum -> Spectrum
    3) 2D Spectrum -> Spectrum
    4) Spectral Cube -> Spectrum
    """
    input_fixture, data_label, expected_cls = data_tuple
    input_data = request.getfixturevalue(input_fixture)
    deconfigged_helper.load(input_data, data_label=data_label, format=data_label)

    # Get actual label from data collection after loading
    label = deconfigged_helper.app.data_collection[0].label
    data = deconfigged_helper.app.data_collection[label]

    # Clear _native_data_cls to force cls inference path
    if '_native_data_cls' in data.meta:
        data.meta.pop('_native_data_cls')

    result = deconfigged_helper._get_data(data_label=label)
    assert isinstance(result, expected_cls)


def test_get_data_cls_spectrum_for_specviz2d(specviz2d_helper, spectrum2d):
    """
    Test _get_data: cls inferred as Spectrum for specviz2d config.
    """
    specviz2d_helper.load_data(spectrum2d)
    # Get the actual label from data collection
    label = specviz2d_helper.app.data_collection[0].label

    # Spectrum should infer Spectrum2d for multi-dimensional data
    result = specviz2d_helper.get_data(data_label=label)
    assert isinstance(result, Spectrum)


def test_get_data_cls_nddataarray_for_rampviz(rampviz_helper, jwst_level_1b_ramp):
    """
    Test _get_data: cls inferred as NDDataArray for rampviz config.
    """
    rampviz_helper.load_data(jwst_level_1b_ramp)
    # Get the actual label from data collection
    label = rampviz_helper.app.data_collection[0].label

    # Rampviz should infer NDDataArray for multi-dimensional data
    result = rampviz_helper.get_data(data_label=label)
    assert isinstance(result, NDDataArray)


def test_delete_region_with_valid_subset(cubeviz_helper, image_cube_hdu_obj):
    """
    Test _delete_region with a valid subset label.
    """
    cubeviz_helper.load(image_cube_hdu_obj, format='3D Spectrum')

    # Create a subset
    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    subset_plugin.import_region(CircularROI(5, 5, 3))

    # Verify subset was created
    subset_labels = [s.label for s in cubeviz_helper.app.data_collection.subset_groups]
    assert 'Subset 1' in subset_labels

    # Delete the region
    cubeviz_helper._delete_region('Subset 1')

    # Verify it was deleted
    subset_labels_after = [s.label for s in cubeviz_helper.app.data_collection.subset_groups]
    assert 'Subset 1' not in subset_labels_after


def test_delete_region_with_invalid_subset(cubeviz_helper, image_cube_hdu_obj):
    """
    Test _delete_region with an invalid subset label (early return).
    """
    cubeviz_helper.load_data(image_cube_hdu_obj)

    # Try to delete a non-existent subset (should not raise error, just return)
    cubeviz_helper._delete_region('NonExistentSubset')

    # Should not raise any errors


def test_load_with_show_in_viewer_deprecated(specviz_helper, spectrum1d):
    """
    Test _load with deprecated 'show_in_viewer' kwarg triggers warning and conversion.
    """
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        specviz_helper.load(spectrum1d, format='1D Spectrum', show_in_viewer=True)
        # Check that deprecation warning was raised
        assert len(w) >= 1
        for warning in w:
            if 'show_in_viewer' in str(warning.message):
                assert issubclass(warning.category, DeprecationWarning)
                break
        else:
            assert False, "DeprecationWarning for 'show_in_viewer' not raised."


def test_load_with_both_viewer_and_show_in_viewer_raises_error(deconfigged_helper, spectrum1d):
    """
    Test _load with both 'viewer' and 'show_in_viewer' raises ValueError.
    """
    with pytest.raises(ValueError, match='Cannot specify both'):
        deconfigged_helper.load(
            spectrum1d, format='1D Spectrum', viewer='spectrum-viewer', show_in_viewer=True)


def test_loaders_not_implemented_without_dev_flag(deconfigged_helper, mosviz_helper):
    """
    Test loaders property raises NotImplementedError without dev flag/or for configs
    not in CONFIGS_WITH_LOADERS.
    """
    # Mosviz is not in CONFIGS_WITH_LOADERS (for now)
    with pytest.raises(NotImplementedError, match='loaders is under active development'):
        _ = mosviz_helper.loaders


def test_get_loader_default_args(deconfigged_helper, spectrum1d):
    """
    Test _get_loader with parser_name is None
    """
    deconfigged_helper.load(spectrum1d, data_label='test', format='1D Spectrum')
    resolver = deconfigged_helper._get_loader('object')
    assert isinstance(resolver, ObjectResolver)

    parser = deconfigged_helper._get_loader('object', parser_name='object')
    assert isinstance(parser, ObjectParser)



from unittest.mock import patch

from specutils import SpectrumList
from jdaviz.core.loaders.importers.importer import BaseImporter
from jdaviz.core.registries import loader_importer_registry
from jdaviz.conftest import _create_spectrum1d_with_spectral_unit
from jdaviz.utils import create_data_hash


# Create a minimal test class that mimics the importer behavior
# without this we get an error when attempting to use BaseImporter directly
class TestBaseImporter(BaseImporter):
    template = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # These are required to be implemented by the BaseImporter class
    # but are currently not tested here
    @property
    def target(self):
        return None

    def __call__(self, *args, **kwargs):
        return None


def test_update_existing_data_in_dc_traitlet(deconfigged_helper, premade_spectrum_list):
    # Test the sync
    test_obj = TestBaseImporter(app=deconfigged_helper.app,
                                resolver=deconfigged_helper.loaders['object']._obj,
                                parser=None,
                                input=premade_spectrum_list)

    assert len(test_obj.existing_data_in_dc) == 0
    assert len(test_obj.app.existing_data_in_dc) == 0

    test_obj.app.existing_data_in_dc.append('test_value')
    # Check that the observer ran and updated the importer traitlet
    assert test_obj.existing_data_in_dc == test_obj.app.existing_data_in_dc

    test_obj.existing_data_in_dc.append('test_value2')
    # The update works on both because of mutability, but it's safer
    # to set the app value directly
    assert test_obj.existing_data_in_dc == test_obj.app.existing_data_in_dc


class TestResetAndCheckExistingDataInDC:
    def test_reset_and_check_basic(self, deconfigged_helper):
        # Can't use a premade_spectrum_list here because the flux data is duplicated
        # and so the data hashes are not unique
        spectrum_list = SpectrumList([_create_spectrum1d_with_spectral_unit(seed=i)
                                      for i in range(5)])

        test_obj = TestBaseImporter(app=deconfigged_helper.app,
                                    resolver=deconfigged_helper.loaders['object']._obj,
                                    parser=None,
                                    input=spectrum_list)

        dh_list = [create_data_hash(spec) for spec in spectrum_list]
        assert len(dh_list) > 0
        # Not populated yet
        assert len(test_obj.existing_data_in_dc) == 0

        # Add some dummy data to existing_data_in_dc
        test_obj.existing_data_in_dc.append('test_value')

        # Now reset using the method
        test_obj.reset_and_check_existing_data_in_dc()
        # existing_data_in_dc should be cleared out and repopulated with any existing data (none)
        assert len(test_obj.existing_data_in_dc) == 0

        # Now load the data into the data collection
        ldr = deconfigged_helper.loaders['object']
        ldr.object = spectrum_list
        ldr.format = '1D Spectrum List'
        ldr.importer.sources.selected = '1D Spectrum at index: 0'
        ldr.load()

        # The data hashes update in the SpectrumList importer but are different from the
        # data hashes in our original test_obj
        dh_list = ldr.importer.sources.data_hashes
        labels_list = ldr.importer.sources.labels
        # The obj would otherwise have all the data hashes/labels from the SpectrumList
        test_obj.data_hashes = dh_list
        test_obj.hash_map_to_label = dict(zip(dh_list, labels_list))

        # These *should* be the same thanks to the observer on the app traitlet
        assert len(test_obj.existing_data_in_dc) == len(test_obj.app.existing_data_in_dc) == 1
        assert dh_list[0] in test_obj.existing_data_in_dc

        # Now reset again, should be the same result
        snackbar_msg = ('Selected data appears to be identical to existing data.\n'
                        '1D Spectrum at index: 0 <=> 1D Spectrum_index-0')  # noqa

        # Mock the broadcast method to catch the snackbar message that will pop out
        # when we run the reset and check again
        with patch.object(deconfigged_helper.app.hub, 'broadcast') as mock_broadcast:
            # TODO: Uncomment when we decide to make this a warning/exception
            # with pytest.warns(UserWarning, match=re.escape(msg)):
            test_obj.reset_and_check_existing_data_in_dc()

            broadcast_msgs = [arg[0][0].text for arg in mock_broadcast.call_args_list
                              if hasattr(arg[0][0], 'text')]
            assert snackbar_msg in broadcast_msgs

        # Everything should be the same as before
        assert len(test_obj.existing_data_in_dc) == 1
        assert dh_list[0] in test_obj.existing_data_in_dc

        new_spectrum_list = SpectrumList([_create_spectrum1d_with_spectral_unit(seed=i)
                                          for i in range(5, 11)])

        # Choose a source at the end to guarantee everything is different from before
        ldr.object = new_spectrum_list
        ldr.format = '1D Spectrum List'
        ldr.importer.sources.selected = '1D Spectrum at index: 5'
        ldr.load()

        dh_list = ldr.importer.sources.data_hashes
        test_obj.data_hashes = dh_list

        # Although two SpectrumList objects are loaded into the data collection,
        # they share no data in common so only one value should be in existing_data_in_dc
        assert len(test_obj.existing_data_in_dc) == len(test_obj.app.existing_data_in_dc) == 1
        assert dh_list[-1] in test_obj.existing_data_in_dc

        deconfigged_helper.app.data_item_remove(test_obj.app.data_collection[1].data.label)
        # Again, existing_data_in_dc is updated from _update_existing_data_in_dc
        # in app.py, however we want to doublecheck that
        # reset_and_check_existing_data_in_dc works as expected
        assert len(test_obj.existing_data_in_dc) == len(deconfigged_helper.app.existing_data_in_dc) == 0  # noqa

        test_obj.reset_and_check_existing_data_in_dc()
        assert len(test_obj.existing_data_in_dc) == len(deconfigged_helper.app.existing_data_in_dc) == 0  # noqa

    def test_reset_and_check_all_importers(self, deconfigged_helper,
                                           image_hdu_wcs, spectrum1d, spectrum2d,
                                           premade_spectrum_list,
                                           sky_coord_only_source_catalog):

        input_data = {'Image': image_hdu_wcs,
                      '1D Spectrum': spectrum1d,
                      '2D Spectrum': spectrum2d,
                      '1D Spectrum List': premade_spectrum_list,
                      '1D Spectrum Concatenated': premade_spectrum_list,
                      'Catalog': sky_coord_only_source_catalog}

        # TODO: Remove when this dev flag is no longer needed
        deconfigged_helper.app.state.catalogs_in_dc = True
        for importer_name, importer in loader_importer_registry.members.items():
            if importer_name not in input_data:
                continue

            # Now load the data into the data collection
            ldr = deconfigged_helper.loaders['object']
            ldr.object = input_data[importer_name]
            ldr.format = importer_name
            ldr.load()

            # Load the same data again
            ldr = deconfigged_helper.loaders['object']
            ldr.object = input_data[importer_name]
            ldr.format = importer_name
            ldr.load()
            test_obj = ldr.importer._obj

            # Mock the broadcast method to catch the snackbar message that should be raised
            # when we run the reset and check again
            with patch.object(deconfigged_helper.app.hub, 'broadcast') as mock_broadcast:
                # TODO: Uncomment when we decide to make this a warning/exception
                # with pytest.warns(UserWarning, match=re.escape(msg)):
                test_obj.reset_and_check_existing_data_in_dc()

                broadcast_msgs = [arg[0][0].text for arg in mock_broadcast.call_args_list
                                  if hasattr(arg[0][0], 'text')]
                assert len(broadcast_msgs) > 0



import numpy as np

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import (BaseImporterToDataCollection,
                                           SpectrumInputExtensionsMixin,
                                           _spectrum_assign_component_type)
from jdaviz.core.user_api import ImporterUserApi


__all__ = ['SpectrumImporter']


@loader_importer_registry('1D Spectrum')
class SpectrumImporter(BaseImporterToDataCollection, SpectrumInputExtensionsMixin):
    template_file = __file__, "./spectrum1d.vue"
    parser_preference = ['fits', 'asdf', 'specutils.Spectrum']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz':
            self.data_label_default = '1D Spectrum'
        elif self.app.config == 'specviz2d':
            self.data_label_default = '1D Spectrum'
        else:
            self.data_label_default = '1D Spectrum'

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '1D Spectrum', 'reference': 'spectrum-1d-viewer'}]

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz', 'specviz2d', 'cubeviz'):
            # NOTE: temporary during deconfig process
            return False
        try:
            if self.spectrum.flux.ndim != 1:
                return False
        except Exception:
            return False
        try:
            self.output
        except Exception:
            return False
        return True

    @property
    def user_api(self):
        expose = ['extension', 'unc_extension', 'mask_extension']
        return ImporterUserApi(self, expose)

    @property
    def supported_flux_ndim(self):
        return 1

    @property
    def output(self):
        # if the entire uncert. array is Nan and the data is not, model fitting won't
        # work (more generally, if uncert[i] is nan/inf and flux[i] is not, fitting will
        # fail, but just deal with the all nan case here since it is straightforward).
        # set uncerts. to None if they are all nan/inf, and display a warning message.
        data = self.spectrum
        if data.uncertainty is not None:
            uncerts_finite = np.isfinite(data.uncertainty.array)
            if not np.any(uncerts_finite):
                data.uncertainty = None
                set_nans_to_none = True

                if set_nans_to_none:
                    # alert user that we have changed their all-nan uncertainty array to None
                    msg = 'All uncertainties are nonfinite, replacing with uncertainty=None.'
                    self.app.hub.broadcast(SnackbarMessage(msg, color="warning", sender=self.app))
        return data

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return _spectrum_assign_component_type(comp_id, comp, units, physical_type)



from traitlets import List, Unicode

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin, SelectPluginComponent
from jdaviz.core.user_api import PluginUserApi

try:
    from jdaviz import __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

__all__ = ['Logger', '_verbosity_levels']

_verbosity_levels = ('debug', 'info', 'warning', 'error')


@tray_registry('logger', label="Logger",
               category='core', sidebar='info', subtab=2)
class Logger(PluginTemplateMixin):
    """Show snackbar messages in a logger UI."""
    template_file = __file__, "logger.vue"

    popup_verbosity_items = List().tag(sync=True)
    popup_verbosity_selected = Unicode('warning').tag(sync=True)
    history_verbosity_items = List().tag(sync=True)
    history_verbosity_selected = Unicode('info').tag(sync=True)

    history = List().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Access history of logger messages.'

        self.popup_verbosity = SelectPluginComponent(self,
                                                     items='popup_verbosity_items',
                                                     selected='popup_verbosity_selected',
                                                     manual_options=_verbosity_levels)

        self.history_verbosity = SelectPluginComponent(self,
                                                       items='history_verbosity_items',
                                                       selected='history_verbosity_selected',
                                                       manual_options=_verbosity_levels)

    @property
    def user_api(self):
        expose = ['popup_verbosity', 'history_verbosity', 'history', 'clear_history']
        return PluginUserApi(self, expose=expose)

    def clear_history(self):
        self.history = []

    def vue_clear_history(self, *args):
        return self.clear_history()

    def queue_message(self, msg, msg_level=None):
        if msg_level not in _verbosity_levels:
            msg_level = 'info'

        msg_level = _verbosity_levels.index(msg_level)
        history_level = _verbosity_levels.index(self.history_verbosity_selected)
        popup_level = _verbosity_levels.index(self.popup_verbosity_selected)

        self.app.state.snackbar_queue.put(self.app.state, self,
                                          msg,
                                          history=msg_level >= history_level,
                                          popup=msg_level >= popup_level)



from traitlets import List, Unicode, observe

from jdaviz.configs.default.plugins.viewers import HistogramViewer
from jdaviz.core.user_api import ViewerCreatorUserApi
from jdaviz.core.viewer_creators import BaseViewerCreator
from jdaviz.core.registries import viewer_creator_registry
from jdaviz.core.template_mixin import SelectPluginComponent
from jdaviz.utils import att_to_componentid


__all__ = ['HistogramViewerCreator']


@viewer_creator_registry('Histogram', overwrite=True)
class HistogramViewerCreator(BaseViewerCreator):
    template_file = __file__, "histogram.vue"
    xatt_items = List().tag(sync=True)
    xatt_selected = Unicode().tag(sync=True)

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.viewer_label_default = 'Histogram'
        self.dataset.filters = ['not_ramp']

        self.xatt = SelectPluginComponent(self,
                                          items='xatt_items',
                                          selected='xatt_selected',
                                          default_mode='first')

    @property
    def user_api(self):
        return ViewerCreatorUserApi(self, expose=('xatt',))

    @property
    def viewer_class(self):
        return HistogramViewer

    @observe('dataset_selected')
    def _on_dataset_selected_change(self, event):
        if not hasattr(self, 'dataset') or not hasattr(self, 'xatt'):
            return
        dc_items = self.dataset.selected_dc_item
        if not self.dataset.is_multiselect:
            dc_items = [dc_items] if dc_items else []
        valid_atts = []
        for dc_item in dc_items:
            valid_atts += [c.label for c in dc_item.components
                           if 'Pixel Axis' not in c.label]
        valid_atts = sorted(set(valid_atts))
        self.xatt._manual_options = valid_atts
        self.xatt._update_items()

    def __call__(self):
        nv = super().__call__()
        gv_state = nv._obj.glue_viewer.state
        if self.xatt.selected != '':
            gv_state.x_att = att_to_componentid(gv_state.x_att_helper, self.xatt.selected)
        if len(nv._obj.glue_viewer.layers):
            nv._obj.glue_viewer.reset_limits()
        return nv



import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import NDData
from regions import CirclePixelRegion, PixCoord, PolygonSkyRegion

from jdaviz.app import Application
from jdaviz.core.config import get_configuration
from jdaviz.core.marks import RegionOverlay
from jdaviz.configs.imviz.helper import Imviz
from jdaviz.configs.imviz.plugins.viewers import ImvizImageView
from jdaviz.configs.imviz.tests.utils import (
    BaseImviz_WCS_NoWCS, create_example_gwcs
)
from jdaviz.conftest import _image_hdu_wcs

from numpy.testing import assert_allclose


@pytest.mark.parametrize(
    ('desired_name', 'actual_name'),
    [(None, 'imviz-1'),
     ('babylon-5', 'babylon-5')])
def test_create_destroy_viewer(imviz_helper, desired_name, actual_name):
    assert imviz_helper.app.get_viewer_ids() == ['imviz-0']

    viewer = imviz_helper.create_image_viewer(viewer_name=desired_name)
    viewer_names = sorted(['imviz-0', actual_name])
    assert viewer.top_visible_data_label == ''
    assert isinstance(viewer, ImvizImageView)
    assert viewer is imviz_helper.app._viewer_store.get(actual_name), list(imviz_helper.app._viewer_store.keys())  # noqa
    assert imviz_helper.app.get_viewer_ids() == viewer_names

    # Make sure plugins that store viewer_items are updated.
    assert sorted(imviz_helper.plugins['Compass'].viewer.labels) == viewer_names

    po = imviz_helper.plugins['Plot Options']
    po.viewer_multiselect = True
    po.layer_multiselect = True
    po.viewer = viewer_names

    imviz_helper.destroy_viewer(actual_name)
    assert imviz_helper.app.get_viewer_ids() == ['imviz-0']
    assert po.viewer.selected == ['imviz-0']
    assert po.viewer.labels == ['imviz-0']


def test_create_viewer_align_by_wcs(imviz_helper, image_2d_wcs):
    data = NDData(np.ones((128, 128)) * u.nJy, wcs=image_2d_wcs)
    imviz_helper.load_data(data, data_label='my_data')

    imviz_helper.create_image_viewer(viewer_name='new-viewer')
    dm = imviz_helper.viewers['new-viewer'].data_menu
    dm.add_data('my_data[DATA]')

    assert not dm._obj.orientation_align_by_wcs

    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    assert dm._obj.orientation_align_by_wcs
    assert dm.orientation.selected == 'Default orientation'


def test_align_by_wcs_create_viewer(imviz_helper, image_2d_wcs):
    data = NDData(np.ones((128, 128)) * u.nJy, wcs=image_2d_wcs)
    imviz_helper.load_data(data, data_label='my_data')

    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    imviz_helper.create_image_viewer(viewer_name='new-viewer')
    dm = imviz_helper.viewers['new-viewer'].data_menu
    dm.add_data('my_data[DATA]')

    assert dm._obj.orientation_align_by_wcs
    assert dm.orientation.selected == 'Default orientation'


def test_get_viewer_created(imviz_helper):
    # This viewer has no reference but has ID.
    viewer1 = imviz_helper.create_image_viewer()
    viewer2 = imviz_helper.app.get_viewer('imviz-1')
    assert viewer1 is viewer2


def test_destroy_viewer_invalid(imviz_helper):
    assert imviz_helper.app.get_viewer_ids() == ['imviz-0']

    imviz_helper.destroy_viewer('foo')
    assert imviz_helper.app.get_viewer_ids() == ['imviz-0']

    with pytest.raises(ValueError, match='cannot be destroyed'):
        imviz_helper.destroy_viewer('imviz-0')
    assert imviz_helper.app.get_viewer_ids() == ['imviz-0']


def test_destroy_viewer_with_subset(imviz_helper):
    """Regression test for https://github.com/spacetelescope/jdaviz/issues/1614"""
    arr = np.ones((10, 10))
    imviz_helper.load_data(arr, data_label='my_array')

    # Create a second viewer.
    imviz_helper.create_image_viewer(viewer_name='second')

    # Add existing data to second viewer.
    imviz_helper.app.add_data_to_viewer('second', 'my_array')

    # Create a Subset.
    reg = CirclePixelRegion(center=PixCoord(x=4, y=4), radius=2)
    imviz_helper.plugins['Subset Tools'].import_region(reg)

    # Remove the second viewer.
    imviz_helper.destroy_viewer('second')

    # Delete the Subset: Should have no traceback.
    imviz_helper.app.delete_subsets('Subset 1')


def test_mastviz_config():
    """Use case from https://github.com/spacetelescope/jdaviz/issues/1037"""

    # create a MAST config dict
    cc = get_configuration('imviz')
    cc['settings']['viewer_spec'] = cc['settings'].get('configuration', 'default')
    cc['settings']['configuration'] = 'mastviz'
    cc['settings']['visible'] = {'menu_bar': False, 'toolbar': False, 'tray': False,
                                 'tab_headers': False}
    cc['toolbar'].remove('g-data-tools') if cc['toolbar'].count('g-data-tools') else None
    cc['toolbar'].remove('g-viewer-creator') if cc['toolbar'].count('g-viewer-creator') else None
    cc['toolbar'].remove('g-image-viewer-creator') if cc['toolbar'].count('g-image-viewer-creator') else None  # noqa

    app = Application(cc)
    im = Imviz(app)
    im.load_data(np.ones((2, 2)), data_label='my_array')

    assert im.app.get_viewer_ids() == ['mastviz-0']
    assert im.app.data_collection[0].shape == (2, 2)


def test_zoom_center_radius_init(imviz_helper):
    """Regression test for https://github.com/spacetelescope/jdaviz/issues/3217"""
    arr = np.ones((10, 10))
    imviz_helper.load_data(arr, data_label='my_array')
    assert imviz_helper.default_viewer._obj.glue_viewer.state.zoom_center_x > 0
    assert imviz_helper.default_viewer._obj.glue_viewer.state.zoom_center_y > 0
    assert imviz_helper.default_viewer._obj.glue_viewer.state.zoom_radius > 0


def test_catalog_in_image_viewer(imviz_helper, image_2d_wcs,
                                 sky_coord_only_source_catalog):
    data = NDData(np.ones((128, 128)) * u.nJy, wcs=image_2d_wcs)
    imviz_helper.load_data(data, data_label='my_data')

    # TODO: remove once dev-flag no longer required
    imviz_helper.app.state.catalogs_in_dc = True

    # change alignment to WCS to enable catalog visibility
    imviz_helper.plugins['Orientation'].align_by = 'WCS'

    # Load the source catalog into the data collection, and into the image viewer
    imviz_helper.load(sky_coord_only_source_catalog, format='Catalog',
                      data_label='my_catalog')

    iv = imviz_helper.viewers['imviz-0']
    dm = iv.data_menu
    po = imviz_helper.plugins['Plot Options']

    # both image and catalog should be in the data menu, able to be
    # loaded/removed from the image viewer
    assert dm.data_labels_visible == ['my_catalog', 'my_data[DATA]']

    assert imviz_helper.app.data_collection[2].label == 'my_catalog'
    assert imviz_helper.app.data_collection[2].meta.get('_importer') == 'CatalogImporter'

    # since catalog is already loaded, it should not be in the "available"
    # choices, but should be in the visible list
    assert 'my_catalog' not in dm._obj.dataset.choices
    assert 'my_catalog' in dm.data_labels_visible

    # should be available from plot options
    assert 'my_catalog' in po.layer.choices

    # test that mouseover appears
    mo = imviz_helper._coords_info
    assert mo.dataset.selected == 'auto'
    assert 'my_catalog' in mo.dataset.choices

    mo._viewer_mouse_event(iv._obj.glue_viewer,
                           {'event': 'mousemove', 'domain': {'x': 0.5, 'y': 0.5}})
    assert mo.as_dict()['data_label'] == 'my_data[DATA]'

    mo.dataset.selected = 'my_catalog'
    mo._viewer_mouse_event(iv._obj.glue_viewer,
                           {'event': 'mousemove', 'domain': {'x': 0.5, 'y': 0.5}})
    assert mo.as_dict()['data_label'] == 'my_catalog'
    assert mo.as_text()[0] == 'Source ID 2'
    assert mo.as_text()[1] == 'World 22h30m06.8256s -20d49m37.4520s (ICRS)'
    # some floating point discrepancies in degs is acceptable, WCS tested elsewhere
    assert len(mo.as_text()[2]) > 0

    # changing to pixel linking should set the layer to hidden
    # and remove from the plot options layer tabs
    imviz_helper.plugins['Orientation'].align_by = 'Pixels'
    assert 'my_catalog' not in dm.data_labels_visible
    assert 'my_catalog' not in po.layer.choices

    # test loading/removing another image dataset, which involves linking
    imviz_helper.load_data(data, data_label='my_data_2')
    dm.layer.selected = ['my_data_2[DATA]']
    dm.remove_from_app()

    # test sucessfully removing the catalog
    dm.layer.selected = ['my_catalog']
    dm.remove_from_app()
    assert 'my_catalog' not in imviz_helper.app.data_collection.labels


def test_get_viewport_sky_region_wcs(imviz_helper, image_hdu_wcs):
    imviz_helper.load(image_hdu_wcs)
    viewer = imviz_helper.viewers['imviz-0']
    region = viewer.get_viewport_region()

    expected_vertices = SkyCoord(
        [337.52124634, 337.52124633, 337.51664042, 337.51664038],
        [-20.83297927, -20.83118685, -20.83118681, -20.83297923],
        unit=u.deg
    )

    assert_allclose(
        region.vertices.separation(expected_vertices).arcsec,
        0, atol=5
    )


def test_get_viewport_sky_region_gwcs(imviz_helper):
    shape = (10, 10)
    data = np.ones(shape)
    ndd = NDData(data=data, wcs=create_example_gwcs(shape))

    imviz_helper.load(ndd)
    viewer = imviz_helper.viewers['imviz-0']
    region = viewer.get_viewport_region()

    expected_vertices = SkyCoord(
        [197.89244966, 197.89271435, 197.89256153, 197.89229684],
        [-1.36574339, -1.36559061, -1.36532599, -1.36547877],
        unit=u.deg
    )

    assert_allclose(
        region.vertices.separation(expected_vertices).arcsec,
        0, atol=5
    )


def test_get_viewport_sky_no_wcs(imviz_helper):
    shape = (10, 10)
    data = np.ones(shape)
    ndd = NDData(data=data)

    imviz_helper.load(ndd)
    viewer = imviz_helper.viewers['imviz-0']

    with pytest.warns(UserWarning, match="does not have valid WCS"):
        viewer.get_viewport_region()


def test_get_viewport_pixel_region(imviz_helper):
    shape = (10, 10)
    data = np.ones(shape)
    ndd = NDData(data=data, wcs=create_example_gwcs(shape))

    imviz_helper.load(ndd)
    viewer = imviz_helper.viewers['imviz-0']

    data_label = imviz_helper.app.data_collection[0].label
    region = viewer.get_viewport_region('pixel', data_label)
    assert_allclose(
        region.vertices.x, [-0.5, -0.5, 9.5, 9.5]
    )
    assert_allclose(
        region.vertices.y, [-0.5, 9.5, 9.5, -0.5]
    )


def test_get_viewport_pixel_region_bad_label(imviz_helper):
    shape = (10, 10)
    data = np.ones(shape)
    ndd = NDData(data=data)
    imviz_helper.load(ndd, data_label='label 1')

    with pytest.raises(ValueError, match="No data found with label xyz"):
        viewer = imviz_helper.viewers['imviz-0']
        viewer.get_viewport_region('pixel', 'xyz')


def test_get_viewport_pixel_region_no_label(imviz_helper):
    shape = (10, 10)
    data = np.ones(shape)
    ndd = NDData(data=data)
    imviz_helper.load(ndd, data_label='label 1')

    with pytest.raises(ValueError,
                       match="`get_viewport_region` requires a data label"):
        viewer = imviz_helper.viewers['imviz-0']
        viewer.get_viewport_region('pixel')


class TestDeleteData(BaseImviz_WCS_NoWCS):

    def test_plot_options_after_destroy(self):
        self.imviz.create_image_viewer(viewer_name="imviz-1")
        self.imviz.app.add_data_to_viewer('imviz-1', 'no_wcs[SCI,1]')

        po = self.imviz.plugins['Plot Options']
        po.open_in_tray()
        po.viewer = "imviz-1"
        po.stretch_function = "Square Root"
        self.imviz.destroy_viewer("imviz-1")
        assert len(po.layer.choices) == 2


class TestRegionOverlay:
    @pytest.fixture(autouse=True)
    def setup_class(self, imviz_helper):
        np.random.seed(42)
        # Data with WCS
        hdu_wcs = _image_hdu_wcs(arr=np.arange(100).reshape((10, 10)))
        imviz_helper.load_data(hdu_wcs)

        self.imviz = imviz_helper
        self.viewer = imviz_helper.default_viewer._obj.glue_viewer

        # Since we are not really displaying, need this to test zoom.
        self.viewer.shape = (100, 100)
        self.viewer.state._set_axes_aspect_ratio(1)

        # create polygon sky regions near the image viewer's sky position:
        ra = np.array([337.51928678, 337.51928679, 337.51697352, 337.5169735])
        dec = np.array([-20.83245874, -20.83160296, -20.83160294, -20.83245871])

        self.regions = []
        self.region_labels = []
        for i in range(5):
            scatter_deg = 2e-4
            ra_offset = np.random.uniform(-scatter_deg, scatter_deg, size=ra.size)
            dec_offset = np.random.uniform(-scatter_deg, scatter_deg, size=ra.size)
            vertices = SkyCoord(ra + ra_offset, dec + dec_offset, unit='deg')
            self.regions.append(PolygonSkyRegion(vertices))
            self.region_labels.append(i)

    def test_add_remove_region_overlay(self):

        self.viewer._add_region_overlay(
            region=self.regions,
            region_label=self.region_labels,
            selected=False
        )

        region_overlay_marks = [
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]

        # check that marks were added:
        n_marks_added = len(region_overlay_marks)
        n_marks_expected = len(self.region_labels)
        assert n_marks_added == n_marks_expected
        assert not any(mark.selected for mark in region_overlay_marks)

        labels_found = [mark.label for mark in region_overlay_marks]
        assert self.region_labels == labels_found

        # remove a region:
        self.viewer._remove_region_overlay(3)
        assert len([
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]) == n_marks_added - 1

        # remove all regions:
        self.viewer._remove_all_region_overlays()
        assert not len([
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ])

    def test_select_region_overlay(self):
        self.viewer._add_region_overlay(
            region=self.regions,
            region_label=self.region_labels,
            selected=False
        )

        select_labels_expected = [0, 1]
        self.viewer._select_region_overlay(
            region_label=select_labels_expected
        )

        region_overlay_marks = [
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]

        # check selected status is correct
        for mark in region_overlay_marks:
            if mark.label in select_labels_expected:
                assert mark.selected
            else:
                assert not mark.selected

        # check that selected regions appear at the end of the
        # marks list, so they're plotted on top of the not-selected marks
        assert select_labels_expected == [mark.label for mark in region_overlay_marks[-2:]]
        assert [2, 3, 4] == [mark.label for mark in region_overlay_marks[:-2]]

        selected_marks = [
            mark for mark in region_overlay_marks if mark.selected
        ]
        not_selected_marks = [
            mark for mark in region_overlay_marks if not mark.selected
        ]

        # check that correct marks are selected
        selected_labels_found = [mark.label for mark in selected_marks]
        assert select_labels_expected == selected_labels_found

        # check that selected and not-selected marks have
        # the expected styles:
        check_selected_attrs = ['stroke_width', 'colors']

        for attr in check_selected_attrs:
            for mark in selected_marks:
                assert getattr(mark, attr) == mark.selected_style[attr]

            for mark in not_selected_marks:
                assert getattr(mark, attr) == mark.default_style[attr]

        # test select all
        self.viewer._select_all_region_overlays()
        n_selected = len([
            mark for mark in self.viewer.figure.marks if mark.selected
        ])
        assert n_selected == len(self.regions)

    def test_deselect_region_overlay(self):
        # all selected
        self.viewer._add_region_overlay(
            region=self.regions,
            region_label=self.region_labels,
            selected=True
        )

        deselect_labels = [0, 3, 4]
        select_labels_expected = [1, 2]
        self.viewer._deselect_region_overlay(
            region_label=deselect_labels
        )

        region_overlay_marks = [
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay)
        ]

        selected_marks = [
            mark for mark in region_overlay_marks if mark.selected
        ]

        # check that correct marks are selected
        selected_labels_found = [mark.label for mark in selected_marks]
        assert select_labels_expected == selected_labels_found

        # deselect all, check none are selected
        self.viewer._deselect_all_region_overlays()
        assert not len([
            mark for mark in self.viewer.figure.marks
            if isinstance(mark, RegionOverlay) and mark.selected
        ])



import numpy as np
import pytest
from astropy import units as u
from specutils import Spectrum


@pytest.mark.filterwarnings('ignore')
def test_linking_after_collapse(cubeviz_helper, spectral_cube_wcs):
    cubeviz_helper.load_data(Spectrum(flux=np.ones((3, 4, 5)) * u.nJy, wcs=spectral_cube_wcs))
    dc = cubeviz_helper.app.data_collection

    # TODO: this now fails when instantiating Collapse after initialization
    coll = cubeviz_helper.plugins['Collapse']._obj

    coll.selected_data_item = '3D Spectrum [FLUX]'
    coll.dataset_selected = '3D Spectrum [FLUX]'

    assert coll.results_label == '3D Spectrum [FLUX] collapsed'
    coll.vue_collapse()
    assert coll.results_label_overwrite is True

    assert len(dc) == 3
    assert dc[2].label == '3D Spectrum [FLUX] collapsed'
    assert len(dc.external_links) == 5

    # Link 3D z to 2D x and 3D y to 2D y

    # Link 1:
    # Pixel Axis 1 [y] from cube.pixel_component_ids[1]
    # Pixel Axis 0 [y] from plugin.pixel_component_ids[0]
    assert dc.external_links[1].cids1[0] == dc[-1].pixel_component_ids[0]
    assert dc.external_links[1].cids2[0] == dc[0].pixel_component_ids[1]
    # Link 2:
    # Pixel Axis 2 [x] from cube.pixel_component_ids[2]
    # Pixel Axis 1 [x] from plugin.pixel_component_ids[1]
    assert dc.external_links[2].cids1[0] == dc[-1].pixel_component_ids[1]
    assert dc.external_links[2].cids2[0] == dc[0].pixel_component_ids[2]


def test_collapse_exception_handling(cubeviz_helper, spectral_cube_wcs):
    cubeviz_helper.load_data(Spectrum(flux=np.ones((3, 4, 5)) * u.nJy, wcs=spectral_cube_wcs))

    coll = cubeviz_helper.plugins['Collapse']._obj

    # Inject bad function selection for use by spectrum.collapse
    # NOTE: Empty strings are allowed per the validating method _selected_changed
    # in SelectPluginComponent... but spectrum.collapse does not accept them
    coll.function.selected = ""
    with pytest.raises(Exception):
        coll.collapse()

    # And following the example from test_linking_after_collapse, we check
    # that the collapse did not in fact proceed
    assert coll.results_label_overwrite is False

    # Now because we've caught the exception when called via vue,
    # there should be no exception raised
    coll.vue_collapse()

    # And check again!
    assert coll.results_label_overwrite is False


def test_collapsed_to_extract_plugin(cubeviz_helper, spectral_cube_wcs, tmp_path):

    cubeviz_helper.load_data(Spectrum(flux=np.ones((3, 4, 5)) * u.nJy, wcs=spectral_cube_wcs))

    collapse_plugin = cubeviz_helper.plugins['Collapse']

    # run collapse function
    collapse_plugin._obj.vue_collapse()

    label = collapse_plugin._obj.add_results.label
    export_plugin = cubeviz_helper.plugins['Export']._obj

    assert label in export_plugin.data_collection.labels



# Licensed under a 3-clause BSD style license - see LICENSE.rst

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''


# Top-level API as exposed to users.
from jdaviz.configs.cubeviz import Cubeviz  # noqa: F401
from jdaviz.configs.imviz import Imviz  # noqa: F401
from jdaviz.configs.mosviz import Mosviz  # noqa: F401
from jdaviz.configs.rampviz import Rampviz  # noqa: F401
from jdaviz.configs.specviz import Specviz  # noqa: F401
from jdaviz.configs.specviz2d import Specviz2d  # noqa: F401
from jdaviz.configs.deconfigged import App  # noqa: F401

from jdaviz.utils import enable_hot_reloading  # noqa: F401
from jdaviz.core.launcher import open  # noqa: F401


_expose = ['show', 'load', 'batch_load',
           'toggle_api_hints',
           'plugins',
           'loaders',
           'viewers',
           'new_viewers',
           'get_data']
_incl = ['enable_hot_reloading', '__version__', 'gca', 'get_all_apps', 'app']
_temporary_incl = ['open', 'Cubeviz', 'Imviz', 'Mosviz', 'Rampviz', 'Specviz', 'Specviz2d']
__all__ = _expose + _incl + _temporary_incl


global _apps
global _current_index

_apps = []


def app(replace=False, set_as_current=True):
    """
    Create a new jdaviz application instance and assign as the current instance.

    Parameters
    ----------
    replace : bool, optional
        If True, replaces the current application instance with the new one.
        Default is False, which means a new instance is added to the list of applications.
    set_as_current : bool, optional
        If True, sets the newly created application instance as the current instance.
        Default is True.

    Returns
    -------
    App
        A new instance of the App class.
    """
    global _apps
    global _current_index
    # NOTE: here we call this "App" for the user, but it is really the "deconfigged"
    # config-helper, which in turn has a .app to access the internal/private Application
    # instance.  After the other configs pass their deprecation period, we should try to
    # rename the internal Application instance and/or merge functionality in with the
    # App class to avoid confusion.
    ca = App(api_hints_obj='jd')
    if replace:
        _apps[_current_index] = ca
    else:
        _apps += [ca]
        if set_as_current:
            _current_index = len(_apps) - 1
    return ca


def get_all_apps():
    """
    Get a list of all jdaviz application instances.

    Returns
    -------
    list
        A list of all jdaviz application instances.
    """
    return _apps


def gca(index=None, set_as_current=True):
    """
    Get the current jdaviz application instance.

    Parameters
    ----------
    index : int, optional
        The index of the application instance to retrieve.
        Default is the current instance.
    set_as_current : bool, optional
        If True, sets the application instance at the specified index as the current instance.
        Default is True.

    Returns
    -------
    App
        The current jdaviz application instance.
    """
    global _current_index
    if not len(_apps):
        # on first access (either to gca() directly or anything redirected through _expose),
        # create a first instance
        app(set_as_current=True)
    if index is None:
        index = _current_index
    _app = _apps[index]
    if set_as_current:
        if index < 0:
            # make sure we have a positive index so
            # this remains fixed when adding new entries
            index = len(_apps) + index
        _current_index = index
    return _app


def __dir__():
    return sorted(__all__)


def __getattr__(name):
    if name in _expose:
        return getattr(gca(), name)
    if name in globals():
        return globals()[name]
    raise AttributeError()
